from __future__ import annotations

import hashlib
import os
import xml.dom.minidom
from typing import Optional, Tuple, List, Any, Coroutine

from meadowrun.abstract_storage_bucket import AbstractStorageBucket
from meadowrun.azure_integration.azure_meadowrun_core import (
    ensure_meadowrun_storage_account,
)
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    MEADOWRUN_RESOURCE_GROUP_NAME,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
    AzureRestApiError,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api import (
    StorageAccount,
    azure_blob_api,
    azure_blob_api_binary,
)

CONTAINER_NAME = "meadowrun"


_LIFECYCLE_POLICY = {
    "rules": [
        {
            "enabled": True,
            "name": "meadowrun-lifecycle-policy",
            "type": "Lifecycle",
            "definition": {
                "actions": {
                    "baseBlob": {"delete": {"daysAfterModificationGreaterThan": 14}}
                },
                "filters": {"blobTypes": ["blockBlob"], "prefixMatch": ["meadowrun/"]},
            },
        }
    ]
}


async def ensure_container(
    subscription_id: str, storage_account: StorageAccount, location: str
) -> None:
    storage_account = await ensure_meadowrun_storage_account(location, "create")
    # https://docs.microsoft.com/en-us/rest/api/storageservices/create-container
    await azure_blob_api(
        "PUT",
        storage_account,
        CONTAINER_NAME,
        query_parameters={"restype": "container"},
        ignored_status_codes=((409, "ContainerAlreadyExists"),),
    )

    await azure_rest_api(
        "PUT",
        f"subscriptions/{subscription_id}/resourceGroups/"
        f"{MEADOWRUN_RESOURCE_GROUP_NAME}/providers/Microsoft.Storage/storageAccounts/"
        f"{storage_account.name}/managementPolicies/default",
        api_version="2021-09-01",
        json_content={"properties": {"policy": _LIFECYCLE_POLICY}},
    )


async def ensure_uploaded(file_path: str) -> str:
    hasher = hashlib.blake2b()
    with open(file_path, "rb") as file:
        buf = file.read()
        hasher.update(buf)
    digest = hasher.hexdigest()

    # the empty location is inelegant here, but:
    # 1. it's only used if the storage account does not exist yet; in that case we're
    #    not even getting to this point
    # 2. it's pretty annoying to thread it where download_file is used.
    storage_account = await ensure_meadowrun_storage_account("", on_missing="raise")
    try:
        _ = await azure_blob_api("HEAD", storage_account, f"{CONTAINER_NAME}/{digest}")
        # CONTAINER_NAME is a constant on Azure, but not AWS. Returning it here for
        # consistency.
        return digest
    except AzureRestApiError as error:
        if error.status != 404:
            raise error

    # doesn't exist, need to upload it
    content_length = os.path.getsize(file_path)
    with open(file_path, "rb") as file:
        _ = await azure_blob_api(
            "PUT",
            storage_account,
            f"{CONTAINER_NAME}/{digest}",
            additional_headers={
                "Content-Length": str(content_length),
                "x-ms-blob-type": "BlockBlob",
            },
            binary_content=file,
        )
    return digest


async def download_file(object_name: str, file_name: str) -> None:
    # see comment in ensure_uploaded_by_hash regarding empty location
    storage_account = await ensure_meadowrun_storage_account("", on_missing="raise")

    result: bytes = await azure_blob_api(
        "GET", storage_account, f"{CONTAINER_NAME}/{object_name}"
    )
    with open(file_name, "wb") as file:
        file.write(result)


async def get_azure_blob_container(location: str) -> AzureBlobContainer:
    return AzureBlobContainer(
        await ensure_meadowrun_storage_account(location, on_missing="create"),
        CONTAINER_NAME,
    )


class AzureBlobContainer(AbstractStorageBucket):
    def __init__(self, storage_account: StorageAccount, container_name: str):
        self._storage_account = storage_account
        self._container_name = container_name
        self._cache_key = f"azure/{storage_account.name}"

    def get_cache_key(self) -> str:
        return self._cache_key

    def get_bytes(self, key: str) -> Coroutine[Any, Any, bytes]:
        return azure_blob_api_binary(
            "GET", self._storage_account, f"{self._container_name}/{key}"
        )

    async def try_get_bytes(self, key: str) -> Optional[bytes]:
        try:
            return await self.get_bytes(key)
        except AzureRestApiError as error:
            if error.status == 404:
                return None
            raise

    def get_byte_range(
        self, key: str, byte_range: Tuple[int, int]
    ) -> Coroutine[Any, Any, bytes]:
        return azure_blob_api_binary(
            "GET",
            self._storage_account,
            f"{self._container_name}/{key}",
            additional_headers={"x-ms-range": f"bytes={byte_range[0]}-{byte_range[1]}"},
        )

    async def write_bytes(self, data: bytes, key: str) -> None:
        await azure_blob_api(
            "PUT",
            self._storage_account,
            f"{self._container_name}/{key}",
            additional_headers={
                "Content-Length": str(len(data)),
                "x-ms-blob-type": "BlockBlob",
            },
            binary_content=data,
        )

    async def exists(self, key: str) -> bool:
        try:
            await azure_blob_api(
                "HEAD", self._storage_account, f"{self._container_name}/{key}"
            )
            return True
        except AzureRestApiError as error:
            if error.status == 404:
                return False
            raise

    async def get_file(self, key: str, local_filename: str) -> None:
        # TODO we should stream this from the response object
        data = await self.get_bytes(key)
        with open(local_filename, "wb") as f:
            f.write(data)

    async def try_get_file(self, key: str, local_filename: str) -> bool:
        try:
            await self.get_file(key, local_filename)
            return True
        except AzureRestApiError as error:
            if error.status == 404:
                return False
            raise

    async def write_file(self, local_filename: str, key: str) -> None:
        content_length = os.path.getsize(local_filename)
        with open(local_filename, "rb") as file:
            _ = await azure_blob_api(
                "PUT",
                self._storage_account,
                f"{self._container_name}/{key}",
                additional_headers={
                    "Content-Length": str(content_length),
                    "x-ms-blob-type": "BlockBlob",
                },
                binary_content=file,
            )

    async def list_objects(self, key_prefix: str) -> List[str]:
        response = await azure_blob_api(
            "GET",
            self._storage_account,
            self._container_name,
            query_parameters={
                "restype": "container",
                "comp": "list",
                "prefix": key_prefix,
            },
        )

        message = xml.dom.minidom.parseString(response)
        root_node = message.documentElement
        if root_node.nodeName != "EnumerationResults":
            raise ValueError("Expected root element to be EnumerationResults")
        results = []
        for node in root_node.childNodes:
            if node.nodeName == "Blobs":
                for blob_node in node.childNodes:
                    for blob_child in blob_node.childNodes:
                        if blob_child.nodeName == "Name":
                            if len(blob_child.childNodes) != 1:
                                raise ValueError("Unexpected XML format")
                            results.append(blob_child.childNodes[0].nodeValue)
        return results

    async def delete_object(self, key: str) -> None:
        await azure_blob_api(
            "DELETE",
            self._storage_account,
            f"{self._container_name}/{key}",
            additional_headers={"x-ms-delete-snapshots": "include"},
        )



from __future__ import annotations

import argparse
import asyncio
import datetime
import os.path
import time
from typing import TYPE_CHECKING

import meadowrun.aws_integration.aws_install_uninstall as aws
import meadowrun.azure_integration.azure_install_uninstall as azure
from meadowrun.aws_integration.aws_core import _get_default_region_name
from meadowrun.aws_integration.aws_permissions_install import (
    grant_permission_to_ecr_repo,
    grant_permission_to_s3_bucket,
    grant_permission_to_secret,
)
from meadowrun.aws_integration.ec2_instance_allocation import SSH_USER
from meadowrun.aws_integration.ec2_pricing import clear_prices_cache
from meadowrun.aws_integration.ec2_ssh_keys import (
    download_ssh_key as ec2_download_ssh_key,
)
from meadowrun.aws_integration.management_lambdas.adjust_ec2_instances import (
    _deregister_and_terminate_instances,
)
from meadowrun.aws_integration.management_lambdas.clean_up import (
    delete_old_task_queues as aws_delete_old_task_queues,
    delete_unused_images as aws_delete_unused_images,
)
from meadowrun.azure_integration.azure_meadowrun_core import (
    ensure_meadowrun_resource_group,
    ensure_meadowrun_storage_account,
    get_default_location,
    get_subscription_id,
)
from meadowrun.azure_integration.azure_ssh_keys import (
    download_ssh_key as azure_download_ssh_key,
    _ensure_meadowrun_vault,
)
from meadowrun.azure_integration.azure_vms import _MEADOWRUN_USERNAME
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    MEADOWRUN_RESOURCE_GROUP_NAME,
    MEADOWRUN_STORAGE_ACCOUNT_KEY_VARIABLE,
    MEADOWRUN_STORAGE_ACCOUNT_VARIABLE,
    MEADOWRUN_SUBSCRIPTION_ID,
)
from meadowrun.azure_integration.mgmt_functions.clean_up import (
    delete_old_task_queues as azure_delete_old_task_queues,
    delete_unused_images as azure_delete_unused_images,
)
from meadowrun.azure_integration.mgmt_functions.vm_adjust import (
    _deregister_and_terminate_vms,
    terminate_all_vms,
)

if TYPE_CHECKING:
    from meadowrun.run_job_core import CloudProviderType


def _strtobool(val: str) -> int:
    # copied from distutils.util
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "True", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "False", "off", "0"):
        return 0
    else:
        raise ValueError("invalid truth value %r" % (val,))


async def async_main(cloud_provider: CloudProviderType) -> None:
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="command")

    # equivalent language for logging
    if cloud_provider == "EC2":
        functions = "AWS lambdas"
        cloud_name = "AWS"
        role = "EC2 role"
        secret = "AWS secret"
        vm_instances = "EC2 instances"
        ssh_user = SSH_USER
        vm_instance_address = "<ec2-instance-public-address>"
        ecr = "ECR"
    elif cloud_provider == "AzureVM":
        functions = "Azure Functions"
        cloud_name = "Azure"
        role = "managed identity"
        secret = "Azure secret"
        vm_instances = "Azure VMs"
        ssh_user = _MEADOWRUN_USERNAME
        vm_instance_address = "<azure-vm-public-ip>"
        ecr = "ACR"
    else:
        raise ValueError(f"Unexpected value for cloud_provider {cloud_provider}")

    install_parser = subparsers.add_parser(
        "install",
        help=f"Does one-time setup of {functions} that automatically periodically "
        "clean up unused temporary resources. Must be re-run when meadowrun is updated "
        f"so that {functions} pick up updated code.",
    )
    install_parser.add_argument(
        "--allow-authorize-ips",
        action="store_true",
        help=(
            "Users' machines need to be authorized to access Meadowrun-created "
            f"{vm_instances}. If this option is set, users will be given permissions "
            "to automatically authorize their IPs to SSH into Meadowrun-created "
            f"{vm_instances}. If this option is not set, an administrator must "
            "manually edit the Meadowrun security group to grant access for users."
        ),
    )
    if cloud_provider == "EC2":
        install_parser.add_argument(
            "--vpc-id",
            type=str,
            help=(
                "An optional argument to specify which VPC to set up the "
                "meadowrun_ssh_security_group in"
            ),
        )
        aws.add_management_lambda_override_arguments(install_parser)

    subparsers.add_parser(
        "uninstall",
        help=f"Removes all {cloud_name} resources created by meadowrun",
    )

    clean_parser = subparsers.add_parser(
        "clean",
        help=f"Cleans up all temporary resources, runs the same code as the {functions}"
        " created by install",
    )
    clean_parser.add_argument("--clean-active", action="store_true")

    if cloud_provider == "EC2":
        grant_permission_to_secret_parser = subparsers.add_parser(
            "grant-permission-to-secret",
            help=f"Gives the meadowrun {role} access to the specified {secret}",
        )
        grant_permission_to_secret_parser.add_argument(
            "secret_name", help=f"The name of the {secret} to give permissions to"
        )

        grant_permission_to_s3_bucket_parser = subparsers.add_parser(
            "grant-permission-to-s3-bucket",
            help=f"Gives the meadowrun {role} access to the specified S3 bucket",
        )
        grant_permission_to_s3_bucket_parser.add_argument(
            "bucket_name", help="The name of the bucket to give permissions to"
        )

        grant_permission_to_ecr_repo_parser = subparsers.add_parser(
            "grant-permission-to-ecr-repo",
            help=f"Gives the meadowrun {role} access to the specified ECR repository",
        )
        grant_permission_to_ecr_repo_parser.add_argument(
            "repo_name", help="The name of the repo to give permissions to"
        )

        edit_management_lambda_config_parser = subparsers.add_parser(
            "edit-management-lambda-config",
            help=(
                "Edits configuration for the management lambdas which are responsible "
                "for cleaning up resources that are no longer in use"
            ),
        )
        aws.add_management_lambda_override_arguments(
            edit_management_lambda_config_parser
        )

    get_ssh_key_parser = subparsers.add_parser(
        "get-ssh-key",
        help="Downloads the SSH key used to connect meadowrun-launched "
        f"{vm_instances}",
    )
    get_ssh_key_parser.add_argument(
        "--output",
        help="The path to write the SSH key to. If it is not provided, the default is "
        "~/.ssh/meadowrun_id_rsa. This can be used with e.g. `ssh -i "
        f"~/.ssh/meadowrun_id_rsa {ssh_user}@{vm_instance_address}`",
    )

    args = parser.parse_args()

    if cloud_provider == "EC2":
        region_name = await _get_default_region_name()
    elif cloud_provider == "AzureVM":
        region_name = get_default_location()
    else:
        raise ValueError(f"Unexpected cloud_provider {cloud_provider}")

    t0 = time.perf_counter()
    if args.command == "install":
        print("Creating resources for running meadowrun")
        if cloud_provider == "EC2":
            await aws.install(
                region_name,
                args.allow_authorize_ips,
                args.vpc_id,
                aws.get_management_lambda_overrides_from_args(args),
            )
        elif cloud_provider == "AzureVM":
            await azure.install(region_name)
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(f"Created resources in {time.perf_counter() - t0:.2f} seconds")
    elif args.command == "edit-management-lambda-config":
        print("Editing management lambda config")
        if cloud_provider == "EC2":
            await aws.edit_management_lambda_config(
                aws.get_management_lambda_overrides_from_args(args)
            )
        elif cloud_provider == "AzureVM":
            raise NotImplementedError(
                "Editing the management lambda config is not yet supported for Azure"
            )
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(
            f"Edited management lambda config in {time.perf_counter() - t0:.2f} seconds"
        )
    elif args.command == "uninstall":
        print("Deleting all meadowrun resources")
        if cloud_provider == "EC2":
            aws.delete_meadowrun_resources(region_name)
        elif cloud_provider == "AzureVM":
            await azure.delete_meadowrun_resource_group()
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(
            f"Deleted all meadowrun resources in {time.perf_counter() - t0:.2f} seconds"
        )
    elif args.command == "clean":
        if args.clean_active:
            print(
                f"Terminating and deregistering all {vm_instances}. This will "
                f"interrupt actively running jobs."
            )
        else:
            print(
                f"Terminating and deregistering all inactive {vm_instances} (specify "
                "--clean-active to also terminate and deregister active instances)"
            )

        if cloud_provider == "EC2":
            if args.clean_active:
                aws.terminate_all_instances(region_name, False)
            _deregister_and_terminate_instances(region_name, datetime.timedelta.min)
            clear_prices_cache()
        elif cloud_provider == "AzureVM":
            resource_group_path = await ensure_meadowrun_resource_group(region_name)
            storage_account = await ensure_meadowrun_storage_account(
                region_name, "raise"
            )
            if args.clean_active:
                await terminate_all_vms(resource_group_path)
            for log_line in await _deregister_and_terminate_vms(
                storage_account, resource_group_path, datetime.timedelta.min
            ):
                print(log_line)
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(
            f"Terminated and deregistered {vm_instances} in "
            f"{time.perf_counter() - t0:.2f} seconds"
        )
        t0 = time.perf_counter()

        print("Deleting unused grid task queues")
        if cloud_provider == "EC2":
            aws_delete_old_task_queues(region_name)
        elif cloud_provider == "AzureVM":
            # set some environment variables as if we're running in an Azure Function
            storage_account = await ensure_meadowrun_storage_account(
                region_name, "create"
            )
            os.environ[MEADOWRUN_STORAGE_ACCOUNT_VARIABLE] = storage_account.name
            os.environ[MEADOWRUN_STORAGE_ACCOUNT_KEY_VARIABLE] = storage_account.key
            os.environ[MEADOWRUN_SUBSCRIPTION_ID] = await get_subscription_id()
            for log_line in await azure_delete_old_task_queues():
                print(log_line)
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(
            f"Deleted unused grid task queues in {time.perf_counter() - t0:.2f} seconds"
        )
        t0 = time.perf_counter()

        print(f"Deleting unused meadowrun-generated {ecr} images")
        if cloud_provider == "EC2":
            aws_delete_unused_images(region_name)
        elif cloud_provider == "AzureVM":
            for log_line in await azure_delete_unused_images():
                print(log_line)
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(
            f"Deleted unused meadowrun-generated {ecr} images in "
            f"{time.perf_counter() - t0:.2f} seconds"
        )
    elif args.command == "grant-permission-to-secret":
        print(f"Granting access to the meadowrun {role} to access {args.secret_name}")
        if cloud_provider == "EC2":
            grant_permission_to_secret(args.secret_name, region_name)
        elif cloud_provider == "AzureVM":
            raise NotImplementedError(
                "Granting permission to individual secrets is not implemented for "
                "Azure. The meadowrun managed identity already has permissions to all "
                "secrets in the meadowrun-created Vault: "
                f"{await _ensure_meadowrun_vault(get_default_location())}"
            )
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(f"Granted access in {time.perf_counter() - t0:.2f} seconds")
    elif args.command == "grant-permission-to-s3-bucket":
        print(f"Granting access to the meadowrun {role} to access {args.bucket_name}")
        if cloud_provider == "EC2":
            grant_permission_to_s3_bucket(args.bucket_name, region_name)
        elif cloud_provider == "AzureVM":
            raise NotImplementedError(
                "Granting permission to S3 buckets is not implemented for Azure. The "
                "meadowrun managed identity already has permissions to all storage "
                f"accounts in the {MEADOWRUN_RESOURCE_GROUP_NAME} resource group"
            )
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(f"Granted access in {time.perf_counter() - t0:.2f} seconds")
    elif args.command == "grant-permission-to-ecr-repo":
        print(f"Granting access to the meadowrun {role} to access {args.repo_name}")
        if cloud_provider == "EC2":
            grant_permission_to_ecr_repo(args.repo_name, region_name)
        elif cloud_provider == "AzureVM":
            raise NotImplementedError(
                "Granting permission to ACR repos is not implemented for Azure. The "
                "meadowrun managed identity already has permissions to all storage "
                f"accounts in the {MEADOWRUN_RESOURCE_GROUP_NAME} resource group"
            )
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
    elif args.command == "get-ssh-key":
        if not args.output:
            output = os.path.expanduser(
                os.path.join("~", ".ssh", f"meadowrun_id_rsa.{cloud_provider}")
            )
        else:
            output = args.output
        print(f"Writing meadowrun ssh key to {output}")
        if cloud_provider == "EC2":
            ec2_download_ssh_key(output, region_name)
        elif cloud_provider == "AzureVM":
            await azure_download_ssh_key(output, region_name)
        else:
            raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
        print(f"Wrote meadowrun ssh key in {time.perf_counter() - t0:.2f} seconds")
    else:
        ValueError(f"Unrecognized command: {args.command}")


def main_ec2() -> None:
    asyncio.run(async_main("EC2"))


def main_azure_vm() -> None:
    asyncio.run(async_main("AzureVM"))



from __future__ import annotations

import abc
import asyncio
import asyncio.subprocess
import dataclasses
import itertools
import os
import os.path
import pathlib
import pickle
import shutil
import struct
import sys
import traceback
from decimal import InvalidOperation
from typing import (
    TYPE_CHECKING,
    Any,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import psutil

from meadowrun.aws_integration.ecr import get_ecr_username_password
from meadowrun.azure_integration.acr import get_acr_username_password
from meadowrun.config import (
    MEADOWRUN_AGENT_PID,
    MEADOWRUN_CODE_MOUNT_LINUX,
    MEADOWRUN_INTERPRETER,
    MEADOWRUN_IO_MOUNT_LINUX,
)
from meadowrun.credentials import (
    CredentialsDict,
    RawCredentials,
    get_docker_credentials,
    get_matching_credentials,
)
from meadowrun.deployment_manager import (
    compile_environment_spec_locally,
    compile_environment_spec_to_container,
    get_code_paths,
)
from meadowrun.docker_controller import (
    get_image_environment_variables_and_working_dir,
    get_registry_domain,
    pull_image,
    remove_container,
    run_container,
)
from meadowrun.meadowrun_pb2 import (
    Credentials,
    Job,
    ProcessState,
    PyAgentJob,
    PyFunctionJob,
    StringPair,
)
from meadowrun.shared import cancel_task, pickle_exception

if TYPE_CHECKING:
    from typing_extensions import Literal

    JobSpecType = Literal["py_command", "py_function", "py_agent"]

    from meadowrun._vendor import aiodocker
    from meadowrun._vendor.aiodocker.containers import DockerContainer
    from meadowrun.deployment_manager import StorageBucketFactoryType
    from meadowrun.run_job_core import CloudProviderType

ProcessStateEnum = ProcessState.ProcessStateEnum

_T = TypeVar("_T")
_U = TypeVar("_U")


_MEADOWRUN_CONTEXT_VARIABLES = "MEADOWRUN_CONTEXT_VARIABLES"
_MEADOWRUN_RESULT_FILE = "MEADOWRUN_RESULT_FILE"
_MEADOWRUN_RESULT_PICKLE_PROTOCOL = "MEADOWRUN_RESULT_PICKLE_PROTOCOL"
MACHINE_CACHE_FOLDER = "/var/meadowrun/machine_cache"


def _string_pairs_to_dict(pairs: Iterable[StringPair]) -> Dict[str, str]:
    """
    Opposite of _string_pairs_from_dict in coordinator_client.py. Helper for dicts in
    protobuf.
    """
    result = {}
    for pair in pairs:
        result[pair.key] = pair.value
    return result


@dataclasses.dataclass
class _JobSpecTransformed:
    """
    To be able to reuse some code, _prepare_py_command and _prepare_py_function compile
    their respective Job.job_specs into a command line that we can run, plus a bit of
    additional information on how to run the command line.
    """

    # the command line to run
    command_line: List[str]

    # This only gets used if we are running in a container. Specifies docker binds to
    # expose files on the host machine for input/output with the container.
    container_binds: List[Tuple[str, str]]

    environment_variables: Dict[str, str] = dataclasses.field(
        default_factory=lambda: {}
    )
    server: Optional[TaskWorkerServer] = None


def _io_file_container_binds(
    io_folder: str, io_files: Iterable[str]
) -> List[Tuple[str, str]]:
    """
    A little helper function. io_folder is a path on the host, io_files are file names
    in that folder, and this function returns binds for docker that map those files into
    the conventional mounts (for meadowrun) within the container.
    """
    return [
        (os.path.join(io_folder, io_file), f"{MEADOWRUN_IO_MOUNT_LINUX}/{io_file}")
        for io_file in io_files
    ]


def _prepare_py_command(
    job: Job, io_folder: str, is_container: bool
) -> _JobSpecTransformed:
    """
    Creates files in io_folder for the child process to use, and returns the
    _JobSpecTransformed
    """

    environment = {}
    io_files = []

    # request the results file
    result_path = os.path.join(io_folder, job.job_id + ".result")
    if is_container:
        result_path_container = f"{MEADOWRUN_IO_MOUNT_LINUX}/{job.job_id}.result"
        # we create an empty file here so that we can expose it to the docker container,
        # docker does not let us bind non-existent files
        open(result_path, "w", encoding="utf-8").close()
    else:
        result_path_container = result_path
    io_files.append(job.job_id + ".result")
    environment[_MEADOWRUN_RESULT_FILE] = result_path_container
    environment[_MEADOWRUN_RESULT_PICKLE_PROTOCOL] = str(
        job.result_highest_pickle_protocol
    )

    # write context variables to file
    if job.py_command.pickled_context_variables:
        context_variables_path = os.path.join(
            io_folder, job.job_id + ".context_variables"
        )
        if is_container:
            context_variables_path_container = (
                f"{MEADOWRUN_IO_MOUNT_LINUX}/{job.job_id}.context_variables"
            )
        else:
            context_variables_path_container = context_variables_path
        with open(context_variables_path, "wb") as f:
            f.write(job.py_command.pickled_context_variables)
        io_files.append(job.job_id + ".context_variables")
        # we can't communicate "directly" with the arbitrary command that the
        # user is running so we'll use environment variables
        environment[_MEADOWRUN_CONTEXT_VARIABLES] = context_variables_path_container

    # get the command line
    if not job.py_command.command_line:
        raise ValueError("command_line must have at least one string")

    return _JobSpecTransformed(
        # we need a list, not a protobuf fake list
        list(job.py_command.command_line),
        _io_file_container_binds(io_folder, io_files),
        environment,
    )


_FUNC_WORKER_PATH = str(
    (
        pathlib.Path(__file__).parent / "func_worker" / "__meadowrun_func_worker.py"
    ).resolve()
)


def _prepare_function(
    job_id: str, function: Union[PyFunctionJob, PyAgentJob], io_folder: str
) -> Tuple[Sequence[str], Sequence[str]]:
    """
    Creates files in io_folder for the child process to use and returns (command line
    arguments, io_files). Compatible with what grid_worker and __meadowrun_func_worker
    expect.
    """
    function_spec = function.WhichOneof("function_spec")
    if function_spec == "qualified_function_name":
        return (
            [
                "--module-name",
                function.qualified_function_name.module_name,
                "--function-name",
                function.qualified_function_name.function_name,
            ],
            [],
        )
    elif function_spec == "pickled_function":
        if function.pickled_function is None:
            raise ValueError("argument cannot be None")
        pickled_function_path = os.path.join(io_folder, job_id + ".function")
        with open(pickled_function_path, "wb") as f:
            f.write(function.pickled_function)
        return ["--has-pickled-function"], [job_id + ".function"]
    else:
        raise ValueError(f"Unknown function_spec {function_spec}")


def _prepare_function_arguments(
    job_id: str, pickled_function_arguments: Optional[bytes], io_folder: str
) -> Tuple[Sequence[str], Sequence[str]]:
    """
    Creates files in io_folder for the child process to use and returns (command line
    arguments, io_files). Compatible with what grid_worker and __meadowrun_func_worker
    expect.
    """

    if pickled_function_arguments:
        pickled_arguments_path = os.path.join(io_folder, job_id + ".arguments")
        with open(pickled_arguments_path, "wb") as f:
            f.write(pickled_function_arguments)
        return ["--has-pickled-arguments"], [job_id + ".arguments"]
    else:
        return [], []


def _prepare_py_function(
    job: Job, io_folder: str, is_container: bool
) -> _JobSpecTransformed:
    """
    Creates files in io_folder for the child process to use and returns
    _JobSpecTransformed. We use __meadowrun_func_worker to start the function in the
    child process.
    """

    io_files = [
        # these line up with __meadowrun_func_worker
        os.path.join(job.job_id + ".state"),
        os.path.join(job.job_id + ".result"),
    ]

    if not is_container:
        func_worker_path = _FUNC_WORKER_PATH
        io_path_container = os.path.join(io_folder, job.job_id)
    else:
        func_worker_path = (
            f"{MEADOWRUN_CODE_MOUNT_LINUX}{os.path.basename(_FUNC_WORKER_PATH)}"
        )
        io_path_container = f"{MEADOWRUN_IO_MOUNT_LINUX}/{job.job_id}"
        for io_file in io_files:
            open(os.path.join(io_folder, io_file), "w", encoding="utf-8").close()

    command_line = [
        "python",
        func_worker_path,
        "--result-highest-pickle-protocol",
        str(job.result_highest_pickle_protocol),
        "--io-path",
        io_path_container,
    ]

    command_line_for_function, io_files_for_function = _prepare_function(
        job.job_id, job.py_function, io_folder
    )

    command_line_for_arguments, io_files_for_arguments = _prepare_function_arguments(
        job.job_id, job.py_function.pickled_function_arguments, io_folder
    )

    return _JobSpecTransformed(
        list(
            itertools.chain(
                command_line, command_line_for_function, command_line_for_arguments
            )
        ),
        _io_file_container_binds(
            io_folder,
            itertools.chain(io_files, io_files_for_function, io_files_for_arguments),
        )
        + [(_FUNC_WORKER_PATH, func_worker_path)],
    )


_TASK_WORKER_PATH = str(
    (
        pathlib.Path(__file__).parent / "func_worker" / "__meadowrun_task_worker.py"
    ).resolve()
)


async def _prepare_py_agent(
    job: Job, io_folder: str, is_container: bool
) -> _JobSpecTransformed:
    """
    Creates files in io_folder for the child process to use and returns
    _JobSpecTransformed. We use __meadowrun_task_worker to start the function in the
    child process.
    """

    io_files = [
        # these line up with __meadowrun_task_worker
        os.path.join(job.job_id + ".state"),
        os.path.join(job.job_id + ".result"),
    ]

    if not is_container:
        worker_path = _TASK_WORKER_PATH
        io_path_container = os.path.join(io_folder, job.job_id)
    else:
        worker_path = (
            f"{MEADOWRUN_CODE_MOUNT_LINUX}{os.path.basename(_TASK_WORKER_PATH)}"
        )
        io_path_container = f"{MEADOWRUN_IO_MOUNT_LINUX}/{job.job_id}"
        for io_file in io_files:
            open(os.path.join(io_folder, io_file), "w", encoding="utf-8").close()

    server = await TaskWorkerServer.start_serving(
        "0.0.0.0" if is_container else "127.0.0.1"
    )
    command_line = [
        "python",
        worker_path,
        "--result-highest-pickle-protocol",
        str(job.result_highest_pickle_protocol),
        "--io-path",
        io_path_container,
        "--host",
        "host.docker.internal" if is_container else "127.0.0.1",
        "--port",
        str(server.port),
    ]

    command_line_for_function, io_files_for_function = _prepare_function(
        job.job_id, job.py_agent, io_folder
    )

    return _JobSpecTransformed(
        list(itertools.chain(command_line, command_line_for_function)),
        _io_file_container_binds(
            io_folder,
            itertools.chain(io_files, io_files_for_function),
        )
        + [(_TASK_WORKER_PATH, worker_path)],
        server=server,
    )


class TaskWorkerServer:
    """This class opens a port in the agent process for the task worker to connect to.
    It allows waiting for connections from task workers, and sending and receiving
    messages.
    """

    def __init__(self) -> None:
        self.server: Optional[asyncio.AbstractServer] = None
        self.reader: Optional[asyncio.StreamReader] = None
        self.writer: Optional[asyncio.StreamWriter] = None
        self.have_connection: asyncio.Event = asyncio.Event()
        self.port: Optional[int] = None

    @staticmethod
    async def start_serving(
        host: str,
    ) -> TaskWorkerServer:
        """Create a TaskWorkerServer, open a free port, and return the server."""
        server = TaskWorkerServer()
        await server._start_serving(host)
        return server

    async def _start_serving(
        self,
        host: str,
    ) -> None:
        server = await asyncio.start_server(self._handle_connection, host, 0)
        (socket,) = server.sockets
        self.port = socket.getsockname()[1]

    async def _handle_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        self.reader = reader
        self.writer = writer
        self.have_connection.set()

    async def wait_for_task_worker_connection(
        self, timeout: float = 30
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Wait until a task worker process connects, with an optional timeout."""
        await asyncio.wait_for(self.have_connection.wait(), timeout=timeout)
        assert self.reader is not None
        assert self.writer is not None
        return self.reader, self.writer

    async def send_message(self, bs: bytes) -> None:
        """Send the given bytes as a message to the task worker. If no task worker is
        connected yet, waits for a connection."""
        if not self.have_connection.is_set():
            await self.wait_for_task_worker_connection()
        assert self.writer is not None
        msg_len = struct.pack(">i", len(bs))
        self.writer.write(msg_len)
        self.writer.write(bs)
        await self.writer.drain()

    async def receive_message(self) -> Any:
        """Receives a message from the task worker. If no task worker is
        connected yet, waits for a connection. The message is unpickled."""
        if not self.have_connection.is_set():
            await self.wait_for_task_worker_connection()
        assert self.reader is not None
        (result_len,) = struct.unpack(">i", await self.reader.read(4))
        result_bs = bytearray()
        while len(result_bs) < result_len:
            result_bs.extend(await self.reader.read(result_len - len(result_bs)))
        return pickle.loads(result_bs)

    async def close_task_worker_connection(self) -> None:
        """Cleanly close the connection to the task worker."""
        if self.writer is not None:
            self.writer.close()
            await self.writer.wait_closed()
            self.have_connection.clear()

    async def close(self) -> None:
        """Cleanly close the connection to the task worker, and stop serving."""
        await self.close_task_worker_connection()
        if self.server:
            self.server.close()
            await self.server.wait_closed()


@dataclasses.dataclass(frozen=True)
class Stats:
    # TODO add cpu usage
    memory_in_use_gb: float


@dataclasses.dataclass
class StatsAccumulator:
    max_memory_used_gb: float = 0

    def accumulate(self, stats: Stats) -> None:
        if stats.memory_in_use_gb > self.max_memory_used_gb:
            self.max_memory_used_gb = stats.memory_in_use_gb

    def snapshot(self) -> Stats:
        return Stats(memory_in_use_gb=self.max_memory_used_gb)


class WorkerMonitor(abc.ABC):
    """Monitors a task worker."""

    def __init__(self) -> None:
        super().__init__()
        self._stats_accumulator = StatsAccumulator()
        self._stats_task: Optional[asyncio.Task[None]] = None

    @abc.abstractmethod
    async def wait_until_exited(self) -> int:
        ...

    @abc.abstractmethod
    async def try_get_return_code(self) -> Optional[int]:
        """
        Returns the exit code of the process if it has already exited, otherwise returns
        None
        """
        ...

    @abc.abstractmethod
    async def restart(self) -> None:
        """Restart the task worker."""
        ...

    @abc.abstractmethod
    async def get_stats(self) -> Stats:
        """Get some stats (memory, cpu...) about the task worker."""
        ...

    async def _start_stats(self) -> None:
        """Subclasses should start this as a task to accumulate statistics,
        and cancel the task when done."""
        self._stats_accumulator = StatsAccumulator()
        while True:
            try:
                stats = await self.get_stats()
                self._stats_accumulator.accumulate(stats)
            except asyncio.CancelledError:
                raise
            except Exception:
                print("Error updating stats:\n" + traceback.format_exc())

            await asyncio.sleep(1)

    def start_stats(self) -> None:
        self._stats_task = asyncio.create_task(self._start_stats())

    async def stop_stats(self) -> StatsAccumulator:
        if self._stats_task is not None:
            await cancel_task(self._stats_task)
        return self._stats_accumulator


class WorkerProcessMonitor(WorkerMonitor):
    """Monitors a task worker that runs as a process."""

    def __init__(
        self,
        command_line: Iterable[str],
        working_directory: Optional[str],
        env_vars: Dict[str, str],
    ):
        super().__init__()
        self.command_line = tuple(command_line)
        self.working_directory = working_directory
        self.env_vars = env_vars
        self._process: Optional[asyncio.subprocess.Process] = None
        self._tail_task: Optional[asyncio.Task[None]] = None

    @property
    def process(self) -> asyncio.subprocess.Process:
        if self._process is None:
            raise InvalidOperation(
                "Worker process not started. Call start_and_tail first."
            )
        return self._process

    @property
    def pid(self) -> int:
        return self.process.pid

    async def start_and_tail(self) -> None:
        self._process = await asyncio.subprocess.create_subprocess_exec(
            *self.command_line,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            cwd=self.working_directory,
            env=self.env_vars,
        )
        self._tail_task = asyncio.create_task(self._tail_log())

    async def stop(self) -> None:
        """Safely stop the process."""
        await self.stop_stats()
        try:
            if self._process is not None:
                self._process.terminate()
                self._process.kill()
        except ProcessLookupError:
            # if it's already gone, we're done here
            pass

    async def wait_until_exited(self) -> int:
        returncode = await self.process.wait()
        if self._tail_task is not None:
            await self._tail_task
        self._tail_task = None
        return returncode

    async def try_get_return_code(self) -> Optional[int]:
        return self.process.returncode

    async def cleanup(self) -> None:
        if self._process is not None:
            await self.stop()
            await self.wait_until_exited()
        if self._tail_task is not None:
            self._tail_task.cancel()
            try:
                await self._tail_task
            except asyncio.CancelledError:
                pass
        self._tail_task = None

    async def _tail_log(self) -> None:
        process = self.process
        assert process.stdout is not None
        async for line in process.stdout:
            sys.stdout.buffer.write(line)

    async def restart(self) -> None:
        await self.stop()
        await self.wait_until_exited()
        await self.start_and_tail()

    async def get_stats(self) -> Stats:
        # TODO cache the psutil process
        process = psutil.Process(self.process.pid)
        # RSS is in bytes: https://man7.org/linux/man-pages/man1/ps.1.html
        return Stats(process.memory_info().rss / (1024**3))


class WorkerContainerMonitor(WorkerMonitor):
    """Monitors a task worker that runs as a container."""

    def __init__(
        self,
        client: Optional[aiodocker.Docker],
        image: str,
        cmd: Optional[List[str]],
        environment_variables: Dict[str, str],
        working_directory: Optional[str],
        binds: List[Tuple[str, str]],
        ports: Iterable[str],
        extra_hosts: List[Tuple[str, str]],
        uses_gpus: bool,
    ):
        self.docker_client = client
        self.image = image
        self.command_line = cmd
        self.env_vars = environment_variables
        self.working_directory = working_directory
        self.binds = binds
        self.ports = list(ports)
        self.extra_hosts = extra_hosts
        self.uses_gpus = uses_gpus
        self._container: Optional[DockerContainer] = None
        self._tail_task: Optional[asyncio.Task[None]] = None

    @property
    def container(self) -> DockerContainer:
        if self._container is None:
            raise InvalidOperation(
                "Container not started yet. Call start_and_tail first."
            )
        return self._container

    @property
    def container_id(self) -> str:
        return self.container.id

    async def start_and_tail(self) -> None:
        self._container, self.docker_client = await run_container(
            self.docker_client,
            self.image,
            self.command_line,
            self.env_vars,
            self.working_directory,
            self.binds,
            self.ports,
            self.extra_hosts,
            self.uses_gpus,
        )
        self._tail_task = asyncio.create_task(self._tail_log())

    async def stop(self) -> None:
        if self._container is not None:
            await self.container.stop()

    async def wait_until_exited(self) -> int:
        wait_result = await self.container.wait()
        if self._tail_task is not None:
            await self._tail_task
        self.tail_task = None
        # as per https://docs.docker.com/engine/api/v1.41/#operation/ContainerWait we
        # can get the return code from the result
        return wait_result["StatusCode"]

    async def try_get_return_code(self) -> Optional[int]:
        container_status = (await self.container.show()).get("State", {})
        if container_status.get("Status", "") == "running":
            # container_status will have ExitCode == 0 while the process is running
            return None
        else:
            return container_status.get("ExitCode", None)

    async def _tail_log(self) -> None:
        # 1. Docker appears to have an objection to having a log driver that can
        # produce plain text files (https://github.com/moby/moby/issues/17020) so we
        # implement that in a hacky way here.
        # TODO figure out overall strategy for logging, maybe eventually implement our
        # own plain text/whatever log driver for docker.
        async for line in self.container.log(stdout=True, stderr=True, follow=True):
            print(f"Task worker: {line}", end="")

    async def restart(self) -> None:
        await self.stop()
        await self.wait_until_exited()
        await self.start_and_tail()

    async def get_stats(self) -> Stats:
        stats = await self.container.stats(stream=False)
        if not stats:
            return Stats(0)
        # according to
        # https://docs.docker.com/engine/api/v1.41/#tag/Container/operation/ContainerStats
        # "memory used" in the docker CLI corresponds to this formula:
        return Stats(
            (
                stats[0]["memory_stats"]["usage"]
                - stats[0]["memory_stats"]["stats"]["cache"]
            )
            / (1024**3)
        )


async def restart_worker(
    server: TaskWorkerServer, worker_monitor: WorkerMonitor
) -> None:
    await server.close_task_worker_connection()
    await worker_monitor.restart()
    await server.wait_for_task_worker_connection()


async def _launch_non_container_job(
    job_spec_type: JobSpecType,
    job_spec_transformed: _JobSpecTransformed,
    code_paths: Sequence[str],
    cwd_path: Optional[str],
    log_file_name: str,
    job: Job,
    io_folder: str,
) -> Tuple[int, Coroutine[Any, Any, ProcessState]]:
    """
    Contains logic specific to launching jobs that run using
    server_available_interpreter. Only separated from _launch_job for readability.
    Assumes that job.server_available_interpreter is populated.

    Returns (pid, continuation), see _launch_job for how to use the continuation.
    """

    # Note that we don't use command_line.io_files here, we only need those for
    # container jobs

    # (1) get the interpreter path:

    interpreter_path = job.server_available_interpreter.interpreter_path
    if interpreter_path == MEADOWRUN_INTERPRETER:
        # replace placeholder
        interpreter_path = sys.executable

    # (2) construct the environment variables dictionary

    env_vars = os.environ.copy()

    # we intentionally overwrite any existing PYTHONPATH--if for some reason we need the
    # current server process' code for the child process, the user needs to include it
    # directly
    env_vars["PYTHONPATH"] = os.pathsep.join(code_paths)

    # On Windows, we believe that interpreter_path can be one of two formats,
    # python_or_venv_dir/python or python_or_venv_dir/Scripts/python. We need to add the
    # scripts directory to the path so that we can run executables as if we're "in the
    # python environment". On Linux, there is no "Scripts" directory and everything is
    # always available in the same directory as the python executable.
    interpreter_path = pathlib.Path(interpreter_path)
    if interpreter_path.parent.name == "Scripts" or os.name != "nt":
        scripts_dir = str(interpreter_path.parent.resolve())
    else:
        scripts_dir = str((interpreter_path.parent / "Scripts").resolve())
    env_vars["PATH"] = scripts_dir + os.pathsep + env_vars["PATH"]

    # Next, merge in env_vars_to_add, computed in the caller. These should take
    # precedence.
    # TODO consider warning if we're overwriting PYTHONPATH or PATH here
    env_vars.update(**job_spec_transformed.environment_variables)

    # (3) get the working directory and fix the command line

    paths_to_search = env_vars["PATH"]
    if cwd_path:
        working_directory = cwd_path
        paths_to_search = f"{working_directory}{os.pathsep}{paths_to_search}"
    else:
        # TODO probably cleanest to allocate a new working directory for each job
        #  instead of just using the default
        working_directory = None

    # Popen uses cwd and env to search for the specified command on Linux but not on
    # Windows according to the docs:
    # https://docs.python.org/3/library/subprocess.html#subprocess.Popen We can use
    # shutil to make the behavior reasonable on both platforms
    new_first_command_line = shutil.which(
        job_spec_transformed.command_line[0], path=paths_to_search
    )
    if new_first_command_line:
        # noinspection PyTypeChecker
        job_spec_transformed.command_line = [
            new_first_command_line
        ] + job_spec_transformed.command_line[1:]

    # (4) run the process

    print(
        f"Running process ({job_spec_type}): "
        f"{' '.join(job_spec_transformed.command_line)}; cwd={working_directory}; "
        f"PYTHONPATH={env_vars['PYTHONPATH']}; log_file_name={log_file_name}; "
        f"code paths={','.join(code_paths)}"
    )

    worker = WorkerProcessMonitor(
        job_spec_transformed.command_line, working_directory, env_vars
    )
    await worker.start_and_tail()

    # (5) return the pid and continuation
    return worker.pid, _non_container_job_continuation(
        worker,
        job_spec_type,
        job_spec_transformed,
        job,
        io_folder,
        log_file_name,
    )


async def _non_container_job_continuation(
    worker: WorkerProcessMonitor,
    job_spec_type: JobSpecType,
    job_spec_transformed: _JobSpecTransformed,
    job: Job,
    io_folder: str,
    log_file_name: str,
) -> ProcessState:
    """
    Takes an asyncio.subprocess.Process, waits for it to finish, gets results from
    io_folder from the child process if necessary, and then returns an appropriate
    ProcessState indicating how the child process completed.
    """

    try:
        if job_spec_type == "py_agent":
            await _run_agent(job_spec_transformed, job, worker)
            await worker.stop()
            await worker.wait_until_exited()
            return ProcessState(
                state=ProcessStateEnum.SUCCEEDED,
                pid=worker.pid or 0,
                log_file_name=log_file_name,
                return_code=0,
            )

        returncode = await worker.wait_until_exited()
        return _completed_job_state(
            job_spec_type,
            job.job_id,
            io_folder,
            log_file_name,
            returncode,
            worker.pid,
            None,
        )
    except asyncio.CancelledError:
        raise
    except Exception as e:
        print(
            "There was an exception while trying to get the final ProcessState:\n"
            + traceback.format_exc()
        )
        return ProcessState(
            state=ProcessStateEnum.ERROR_GETTING_STATE,
            pickled_result=pickle_exception(e, job.result_highest_pickle_protocol),
        )
    finally:
        # TODO this is not 100% bulletproof--there will be a tiny sliver of time between
        # when we create the process and we enter this try/finally block
        try:
            await worker.stop()
        except BaseException:
            print("Exception trying to kill process: " + traceback.format_exc())


async def _launch_container_job(
    job_spec_type: JobSpecType,
    container_image_name: str,
    job_spec_transformed: _JobSpecTransformed,
    code_paths: Sequence[str],
    cwd_path: Optional[str],
    log_file_name: str,
    job: Job,
    io_folder: str,
    sidecar_container_images: List[str],
) -> Tuple[str, Coroutine[Any, Any, ProcessState]]:
    """
    Contains logic specific to launching jobs that run in a container. Only separated
    from _launch_job for readability.

    job_spec_transformed.environment_variables will take precedence over any environment
    variables specified in the container.

    Assumes that the container image has been pulled to this machine already.

    Returns (container_id, continuation), see _launch_job for how to use the
    continuation.
    """

    binds: List[Tuple[str, str]] = [(MACHINE_CACHE_FOLDER, MACHINE_CACHE_FOLDER)]

    if cwd_path is not None:
        unique_code_paths = list(dict.fromkeys(itertools.chain([cwd_path], code_paths)))
    else:
        unique_code_paths = list(code_paths)

    # populate binds with code paths we need
    # TODO this isn't exactly right--some code paths "overlap", like foo/a and foo/a/b.
    # We should just bind foo/a once as e.g. /meadowrun/code0, and then add both
    # /meadowrun/code0 and /meadowrun/code0/b to the PYTHONPATH. Instead, right now, we
    # mount /foo/a as /meadowrun/code0 and /foo/a/b as /meadowrun/code1, which might
    # lead to unexpected behavior.
    for i, path_on_host in enumerate(unique_code_paths):
        mounted_code_path = f"{MEADOWRUN_CODE_MOUNT_LINUX}{i}"
        binds.append((path_on_host, mounted_code_path))

    # get the image's environment variables and working directory
    (
        image_environment_variables,
        working_dir,
    ) = await get_image_environment_variables_and_working_dir(container_image_name)

    # If the container image has a meaningful working directory set, then we want to
    # leave that as it is. But if the working directory is "/tmp/", we use that as a
    # placeholder to mean that we don't care about the container image's working
    # directory. In that case, we set it to cwd_path, if given.
    if working_dir in ("/tmp/", "/tmp") and cwd_path is not None:
        (working_dir,) = tuple(
            mounted_path for (host_path, mounted_path) in binds if host_path == cwd_path
        )

    # If we've exposed any code paths, add them to PYTHONPATH. The normal behavior for
    # environment variables is that if they're specified in job_spec_transformed, those
    # override (rather than append to) what's defined in the container. If they aren't
    # specified in job_spec_transformed, we use whatever is specified in the container
    # image. We need to replicate that logic here so that we just add mounted_code_paths
    # to whatever PYTHONPATH would have "normally" become.
    if len(binds) > 1:
        existing_python_path = []
        if "PYTHONPATH" in job_spec_transformed.environment_variables:
            existing_python_path = [
                job_spec_transformed.environment_variables["PYTHONPATH"]
            ]
        else:
            if image_environment_variables:
                for image_env_var in image_environment_variables:
                    if image_env_var.startswith("PYTHONPATH="):
                        existing_python_path = [image_env_var[len("PYTHONPATH=") :]]
                        # just take the first PYTHONPATH we see, not worth worrying
                        # about pathological case where there are multiple PYTHONPATH
                        # environment variables
                        break

        code_paths = tuple(
            mounted_path
            for (host_path, mounted_path) in binds
            if host_path in code_paths
        )
        job_spec_transformed.environment_variables["PYTHONPATH"] = os.pathsep.join(
            itertools.chain(code_paths, existing_python_path)
        )

    # now, expose any files we need for communication with the container
    binds.extend(job_spec_transformed.container_binds)

    # now run any sidecar_containers that were specified
    docker_client = None

    sidecar_container_ips = []
    sidecar_containers = []
    for sidecar_container_image in sidecar_container_images:
        sidecar_container, docker_client = await run_container(
            docker_client,
            sidecar_container_image,
            None,
            {},
            None,
            [],
            [],
            [],
            job.uses_gpu,
        )
        sidecar_containers.append(sidecar_container)
        sidecar_container_ips.append(
            (await sidecar_container.show())["NetworkSettings"]["IPAddress"]
        )

    # finally, run the container
    print(
        f"Running container ({job_spec_type}): "
        f"{' '.join(job_spec_transformed.command_line)}; "
        f"container image={container_image_name}; "
        f"PYTHONPATH={job_spec_transformed.environment_variables.get('PYTHONPATH')} "
        f"log_file_name={log_file_name}; code paths={','.join(code_paths)} "
        f"ports={','.join(port for port in job.ports)}"
    )

    worker = WorkerContainerMonitor(
        docker_client,
        container_image_name,
        # json serializer needs a real list, not a protobuf fake list
        job_spec_transformed.command_line,
        job_spec_transformed.environment_variables,
        working_dir,
        binds,
        list(job.ports),
        [(f"sidecar-container-{i}", ip) for i, ip in enumerate(sidecar_container_ips)],
        job.uses_gpu,
    )
    await worker.start_and_tail()

    return worker.container_id, _container_job_continuation(
        worker,
        job_spec_type,
        job_spec_transformed,
        job,
        io_folder,
        log_file_name,
        sidecar_containers,
    )


async def _container_job_continuation(
    worker: WorkerContainerMonitor,
    job_spec_type: JobSpecType,
    job_spec_transformed: _JobSpecTransformed,
    job: Job,
    io_folder: str,
    log_file_name: str,
    sidecar_containers: List[DockerContainer],
) -> ProcessState:
    """
    Writes the container's logs to log_file_name, waits for the container to finish, and
    then returns a ProcessState indicating the state of this container when it
    finished.

    docker_client just needs to be closed when the container process has completed.
    """
    try:
        if job_spec_type == "py_agent":
            await _run_agent(job_spec_transformed, job, worker)
            await worker.stop()
            await worker.wait_until_exited()
            return ProcessState(
                state=ProcessStateEnum.SUCCEEDED,
                container_id=worker.container_id or "",
                log_file_name=log_file_name,
                return_code=0,
            )

        return_code = await worker.wait_until_exited()
        return _completed_job_state(
            job_spec_type,
            job.job_id,
            io_folder,
            log_file_name,
            return_code,
            None,
            worker.container_id,
        )
    except asyncio.CancelledError:
        raise
    except Exception as e:
        print(
            "There was an exception while trying to get the final ProcessState:\n"
            + traceback.format_exc()
        )
        return ProcessState(
            state=ProcessStateEnum.ERROR_GETTING_STATE,
            pickled_result=pickle_exception(e, job.result_highest_pickle_protocol),
        )
    finally:
        # TODO this is not 100% bulletproof--there will be a tiny sliver of time between
        # when we create the container and we enter this try/finally block
        try:
            await asyncio.gather(
                remove_container(worker.container),
                *(remove_container(c) for c in sidecar_containers),
            )
        except BaseException:
            print("Warning, unable to remove container: " + traceback.format_exc())
        assert worker.docker_client is not None
        await worker.docker_client.__aexit__(None, None, None)


async def _run_agent(
    job_spec_transformed: _JobSpecTransformed, job: Job, worker: WorkerMonitor
) -> None:
    # run the agent function. The agent function connects to another
    # process, the task worker, which runs the actual user function.
    assert job_spec_transformed.server is not None
    agent_func = pickle.loads(job.py_agent.pickled_agent_function)
    agent_func_args, agent_func_kwargs = pickle.loads(
        job.py_agent.pickled_agent_function_arguments
    )
    await agent_func(
        *agent_func_args,
        worker_server=job_spec_transformed.server,
        worker_monitor=worker,
        **agent_func_kwargs,
    )
    await job_spec_transformed.server.close()


def _completed_job_state(
    job_spec_type: JobSpecType,
    job_id: str,
    io_folder: str,
    log_file_name: str,
    return_code: int,
    pid: Optional[int],
    container_id: Optional[str],
) -> ProcessState:
    """
    This creates an appropriate ProcessState for a job that has completed (regardless of
    whether it ran in a process or container).
    """

    # see if we got a normal return code
    if return_code != 0:
        return ProcessState(
            state=ProcessStateEnum.NON_ZERO_RETURN_CODE,
            # TODO some other number? we should have either pid or container_id.
            pid=pid or 0,
            container_id=container_id or "",
            log_file_name=log_file_name,
            return_code=return_code,
        )

    # if we returned normally

    # for py_funcs or py_agents, get the state
    if job_spec_type in ["py_function", "py_agent"]:
        state_file = os.path.join(io_folder, job_id + ".state")
        with open(state_file, "r", encoding="utf-8") as state_file_text_reader:
            state_string = state_file_text_reader.read()
        if state_string == "SUCCEEDED":
            state = ProcessStateEnum.SUCCEEDED
        elif state_string == "PYTHON_EXCEPTION":
            state = ProcessStateEnum.PYTHON_EXCEPTION
        else:
            raise ValueError(f"Unknown state string: {state_string}")
    elif job_spec_type == "py_command":
        state = ProcessStateEnum.SUCCEEDED
    else:
        raise ValueError(f"job_spec was not recognized {job_spec_type}")

    # Next get the result. The result file is optional for py_commands because we don't
    # have full control over the process and there's no way to guarantee that "our code"
    # gets executed
    result_file = os.path.join(io_folder, job_id + ".result")
    if job_spec_type != "py_command" or os.path.exists(result_file):
        with open(result_file, "rb") as result_file_reader:
            result = result_file_reader.read()
    else:
        result = b""

    # TODO clean up files in io_folder for this process

    # return the ProcessState
    return ProcessState(
        state=state,
        pid=pid or 0,
        container_id=container_id or "",
        log_file_name=log_file_name,
        return_code=0,
        pickled_result=result,
    )


def _set_up_working_folder() -> Tuple[str, str, str, str, str]:
    """
    Returns conventional io, job_logs, git_repos, local_copies, and misc path names.
    It's assumed that these paths already exist.
    """

    working_folder = "/var/meadowrun"

    # this holds files for transferring data to and from this server process and the
    # child processes
    io_folder = os.path.join(working_folder, "io")
    # holds the logs for the functions/commands that this server runs
    job_logs_folder = os.path.join(working_folder, "job_logs")
    # see CodeDeploymentManager
    git_repos_folder = os.path.join(working_folder, "git_repos")
    # see CodeDeploymentManager
    local_copies_folder = os.path.join(working_folder, "local_copies")
    # misc folder for e.g. storing environment export files sent from local machine
    misc_folder = os.path.join(working_folder, "misc")

    return (
        io_folder,
        job_logs_folder,
        git_repos_folder,
        local_copies_folder,
        misc_folder,
    )


def _get_credentials_sources(job: Job) -> CredentialsDict:
    credentials_sources: CredentialsDict = {}
    for credentials_source in job.credentials_sources:
        source = credentials_source.WhichOneof("source")
        if source is None:
            raise ValueError(
                "CredentialsSourceMessage should have a source set: "
                f"{credentials_source}"
            )
        credentials_sources.setdefault(credentials_source.service, []).append(
            (credentials_source.service_url, getattr(credentials_source, source))
        )

    return credentials_sources


async def _get_credentials_for_docker(
    repository: str,
    credentials_sources: CredentialsDict,
    cloud: Optional[Tuple[CloudProviderType, str]],
) -> Optional[RawCredentials]:
    """
    Tries to get the credentials for a docker repository from credentials_sources. If
    that doesn't succeed, try to get a username/password for ECR/ACR based on the
    current role.
    """
    result = None
    try:
        result = await get_docker_credentials(repository, credentials_sources)
    except asyncio.CancelledError:
        raise
    except Exception:
        print("Error trying to turn credentials source into actual credentials")
        traceback.print_exc()

    if result is None and cloud is not None:
        try:
            registry_domain, _ = get_registry_domain(repository)
            if cloud[0] == "EC2":
                result = get_ecr_username_password(registry_domain)
            elif cloud[0] == "AzureVM":
                result = await get_acr_username_password(registry_domain)
            else:
                raise ValueError("Unexpected value for CloudProviderType: {cloud[0]}")
        except asyncio.CancelledError:
            raise
        except Exception:
            print(f"Error trying to get cloud credentials for {repository}")
            traceback.print_exc()

    return result


async def _get_credentials_for_job(
    job: Job,
    cloud: Optional[Tuple[CloudProviderType, str]],
) -> Tuple[
    Optional[RawCredentials], Optional[RawCredentials], List[Optional[RawCredentials]]
]:
    """
    Returns (credentials for job.code_deployment, credentials for
    job.interpreter_deployment, credentials for job.sidecar_containers). Credentials for
    sidecar_containers will be a list of the same length as sidecar_containers, and
    should be used like zip(job.sidecar_containers, returned_value).

    TODO: This code seems a little convoluted because on the client side, the user
    specifies which credentials are for which deployment/sidecar container. We should
    change this code to reflect the user's specifications, but the code we have here is
    also useful if we want to support registering "global" credentials that are stored
    outside of the context of a single job.
    """

    # first, get all available credentials sources from the JobToRun
    credentials_sources = _get_credentials_sources(job)

    # now, get any matching credentials sources and turn them into credentials
    code_deployment_credentials, interpreter_deployment_credentials = None, None

    code_deployment_type = job.WhichOneof("code_deployment")
    if code_deployment_type in ("git_repo_commit", "git_repo_branch"):
        if code_deployment_type == "git_repo_commit":
            repo_url = job.git_repo_commit.repo_url
        else:
            repo_url = job.git_repo_branch.repo_url

        try:
            code_deployment_credentials = await get_matching_credentials(
                Credentials.Service.GIT, repo_url, credentials_sources
            )
        except asyncio.CancelledError:
            raise
        except Exception:
            # TODO ideally this would make it back to an error message for job if it
            #  eventually fails (and maybe even if it doesn't)
            print("Error trying to turn credentials source into actual credentials")
            traceback.print_exc()

    interpreter_deployment_type = job.WhichOneof("interpreter_deployment")
    if interpreter_deployment_type in ("container_at_digest", "container_at_tag"):
        if interpreter_deployment_type == "container_at_digest":
            repository = job.container_at_digest.repository
        else:
            repository = job.container_at_tag.repository

        interpreter_deployment_credentials = await _get_credentials_for_docker(
            repository, credentials_sources, cloud
        )

    sidecar_container_credentials: List[Optional[RawCredentials]] = []
    for sidecar_container in job.sidecar_containers:
        sidecar_container_image = sidecar_container.WhichOneof("container_image")

        if sidecar_container_image in (
            "container_image_at_digest",
            "container_image_at_tag",
        ):
            if sidecar_container_image == "container_image_at_digest":
                repository = sidecar_container.container_image_at_digest.repository
            else:
                repository = sidecar_container.container_image_at_tag.repository

            sidecar_container_credentials.append(
                await _get_credentials_for_docker(
                    repository, credentials_sources, cloud
                )
            )
        else:
            sidecar_container_credentials.append(None)

    return (
        code_deployment_credentials,
        interpreter_deployment_credentials,
        sidecar_container_credentials,
    )


async def run_local(
    job: Job,
    cloud: Optional[Tuple[CloudProviderType, str]] = None,
    storage_bucket_factory: StorageBucketFactoryType = None,
    compile_environment_in_container: bool = True,
) -> Tuple[ProcessState, Optional[asyncio.Task[ProcessState]]]:
    """
    Runs a job locally using the specified working_folder (or uses the default). Meant
    to be called on the "server" where the client is calling e.g. run_function.

    Returns a tuple of (initial job state, continuation).

    The initial ProcessState will either be RUNNING or RUN_REQUEST_FAILED. If the
    initial job state is RUNNING, this should be reported back to the client so that the
    user knows the pid and log_file_name of the child process for that job.

    If the initial job state is RUNNING, then continuation will be an asyncio Task that
    will complete once the child process has completed, and then return another
    ProcessState that indicates how the job completed, e.g. SUCCEEDED, PYTHON_EXCEPTION,
    NON_ZERO_RETURN_CODE.

    If the initial job state is RUN_REQUEST_FAILED, the continuation will be None.

    compile_environment_in_container controls the behavior when job has an environment
    spec for the interpreter. If compile_environment_in_container is true, environment
    specs get compiled into container images. If it's false, environment specs are
    turned into environments in the local environment
    TODO compile_environment_in_container should probably be part of the Job object.
    """

    (
        io_folder,
        job_logs_folder,
        git_repos_folder,
        local_copies_folder,
        misc_folder,
    ) = _set_up_working_folder()

    # the logging actually happens via stdout redirection in the run_job_local_main
    # caller
    log_file_name = os.path.join(
        job_logs_folder,
        f"{job.job_friendly_name}.{job.job_id}.log",
    )

    try:
        # unpickle credentials if necessary
        (
            code_deployment_credentials,
            interpreter_deployment_credentials,
            all_sidecar_container_credentials,
        ) = await _get_credentials_for_job(job, cloud)

        # first, get the code paths
        code_paths, interpreter_spec_path, cwd_path = await get_code_paths(
            git_repos_folder,
            local_copies_folder,
            job,
            code_deployment_credentials,
            storage_bucket_factory,
        )

        # next, if we have a environment_spec_in_code, turn into a container

        interpreter_deployment = job.WhichOneof("interpreter_deployment")

        if interpreter_deployment == "environment_spec_in_code":
            if interpreter_spec_path is None:
                raise ValueError(
                    "Cannot specify environment_spec_in_code and not provide any code "
                    "paths"
                )
            if compile_environment_in_container:
                job.server_available_container.CopyFrom(
                    await compile_environment_spec_to_container(
                        job.environment_spec_in_code, interpreter_spec_path, cloud
                    )
                )
                interpreter_deployment = "server_available_container"
            else:
                job.server_available_interpreter.CopyFrom(
                    await compile_environment_spec_locally(
                        job.environment_spec_in_code, interpreter_spec_path, misc_folder
                    )
                )
                interpreter_deployment = "server_available_interpreter"

        if interpreter_deployment == "environment_spec":
            if compile_environment_in_container:
                job.server_available_container.CopyFrom(
                    await compile_environment_spec_to_container(
                        job.environment_spec, misc_folder, cloud
                    )
                )
                interpreter_deployment = "server_available_container"
            else:
                job.server_available_interpreter.CopyFrom(
                    await compile_environment_spec_locally(
                        job.environment_spec, misc_folder, misc_folder
                    )
                )
                interpreter_deployment = "server_available_interpreter"

        # then decide if we're running in a container or not

        is_container = interpreter_deployment in (
            "container_at_digest",
            "container_at_tag",
            "server_available_container",
        )

        # next, transform job_spec into _JobSpecTransformed. _JobSpecTransformed can all
        # be run the same way, i.e. we no longer have to worry about the differences in
        # the job_specs after this section

        job_spec_type = job.WhichOneof("job_spec")

        if job_spec_type == "py_command":
            job_spec_transformed = _prepare_py_command(job, io_folder, is_container)
        elif job_spec_type == "py_function":
            job_spec_transformed = _prepare_py_function(job, io_folder, is_container)
        elif job_spec_type == "py_agent":
            job_spec_transformed = await _prepare_py_agent(job, io_folder, is_container)
        else:
            raise ValueError(f"Unknown job_spec {job_spec_type}")

        # next, prepare a few other things

        # add the agent pid, but don't modify if it already exists somehow
        if MEADOWRUN_AGENT_PID not in job_spec_transformed.environment_variables:
            job_spec_transformed.environment_variables[MEADOWRUN_AGENT_PID] = str(
                os.getpid()
            )

        # add PYTHONUNBUFFERED=1, but don't modify if it already exists
        if "PYTHONUNBUFFERED" not in job_spec_transformed.environment_variables:
            job_spec_transformed.environment_variables["PYTHONUNBUFFERED"] = "1"

        # Merge in the user specified environment, variables, these should always take
        # precedence. A little sloppy to modify in place, but should be fine
        # TODO consider warning if we're overwriting any variables that already exist
        job_spec_transformed.environment_variables.update(
            **_string_pairs_to_dict(job.environment_variables)
        )

        # next we need to launch the job depending on how we've specified the
        # interpreter

        if interpreter_deployment == "server_available_interpreter":
            if job.sidecar_containers:
                raise ValueError(
                    "Cannot specify sidecar_containers with "
                    "server_available_interpreter"
                )
            pid, continuation = await _launch_non_container_job(
                job_spec_type,
                job_spec_transformed,
                code_paths,
                cwd_path,
                log_file_name,
                job,
                io_folder,
            )
            # due to the way protobuf works, this is equivalent to None
            container_id = ""
        elif is_container:
            if interpreter_deployment == "container_at_digest":
                container_image_name = (
                    f"{job.container_at_digest.repository}"
                    f"@{job.container_at_digest.digest}"
                )
                await pull_image(
                    container_image_name, interpreter_deployment_credentials
                )
            elif interpreter_deployment == "container_at_tag":
                # warning this is not reproducible!!! should ideally be resolved on the
                # client
                container_image_name = (
                    f"{job.container_at_tag.repository}:{job.container_at_tag.tag}"
                )
                await pull_image(
                    container_image_name, interpreter_deployment_credentials
                )
            elif interpreter_deployment == "server_available_container":
                container_image_name = job.server_available_container.image_name
                # server_available_container assumes that we do not need to pull, and it
                # may not be possible to pull it (i.e. it only exists locally)
            else:
                raise ValueError(
                    f"Unexpected interpreter_deployment: {interpreter_deployment}"
                )

            sidecar_containers = []
            for sidecar_container, sidecar_container_credentials in zip(
                job.sidecar_containers, all_sidecar_container_credentials
            ):
                sidecar_container_image_type = sidecar_container.WhichOneof(
                    "container_image"
                )
                if sidecar_container_image_type == "container_image_at_digest":
                    sidecar_containers.append(
                        f"{sidecar_container.container_image_at_digest.repository}"
                        f"@{sidecar_container.container_image_at_digest.digest}"
                    )
                    await pull_image(
                        sidecar_containers[-1], sidecar_container_credentials
                    )
                elif sidecar_container_image_type == "container_image_at_tag":
                    sidecar_containers.append(
                        f"{sidecar_container.container_image_at_tag.repository}"
                        f":{sidecar_container.container_image_at_tag.tag}"
                    )
                    await pull_image(
                        sidecar_containers[-1], sidecar_container_credentials
                    )
                elif sidecar_container_image_type == "server_available_container_image":
                    sidecar_containers.append(
                        sidecar_container.server_available_container_image.image_name
                    )
                else:
                    raise ValueError(
                        "Unexpected sidecar container image: "
                        f"{sidecar_container_image_type}"
                    )

            container_id, continuation = await _launch_container_job(
                job_spec_type,
                container_image_name,
                job_spec_transformed,
                code_paths,
                cwd_path,
                log_file_name,
                job,
                io_folder,
                sidecar_containers,
            )
            # due to the way protobuf works, this is equivalent to None
            pid = 0
        else:
            raise ValueError(
                f"Did not recognize interpreter_deployment {interpreter_deployment}"
            )

        # launching the process succeeded, return the RUNNING state and create the
        # continuation
        return (
            ProcessState(
                state=ProcessStateEnum.RUNNING,
                pid=pid,
                container_id=container_id,
                log_file_name=log_file_name,
            ),
            asyncio.create_task(continuation),
        )
    except asyncio.CancelledError:
        raise
    except Exception as e:
        traceback.print_exc()

        # we failed to launch the process
        return (
            ProcessState(
                state=ProcessStateEnum.RUN_REQUEST_FAILED,
                pickled_result=pickle_exception(e, job.result_highest_pickle_protocol),
                log_file_name=log_file_name,
            ),
            None,
        )



from __future__ import annotations

import dataclasses
import decimal
import math
import sys
from typing import Any, Dict, Optional, Tuple, Iterable, Union

from typing_extensions import Literal

from meadowrun.config import (
    AVX,
    GPU,
    GPU_MEMORY,
    EVICTION_RATE_INVERSE,
    LOGICAL_CPU,
    MEMORY_GB,
    avx_from_string,
)

ON_DEMAND_OR_SPOT_VALUES: Tuple[OnDemandOrSpotType, OnDemandOrSpotType] = (
    "on_demand",
    "spot",
)
OnDemandOrSpotType = Literal["on_demand", "spot"]


class ResourcesInternal:
    """
    Can represent both resources available (i.e. on a agent) as well as resources
    required (i.e. by a job).

    This is called ResourcesInternal to distinguish it from Resources. Resources is
    mostly used as a friendly way to ultimately construct a ResourcesInternal object,
    which is used internally. Only Resources should be part of public APIs.
    """

    def __init__(self, consumable: Dict[str, float], non_consumable: Dict[str, float]):
        self.consumable = consumable
        self.non_consumable = non_consumable

    def subtract(self, required: ResourcesInternal) -> Optional[ResourcesInternal]:
        """
        Interpreting self as "resources available" on an instance, subtracts "resources
        required" for a job on this instance

        Returns None if the required resources are not available in self.
        """

        for key, value in required.non_consumable.items():
            if key not in self.non_consumable:
                return None
            if self.non_consumable[key] < required.non_consumable[key]:
                return None

        for key, value in required.consumable.items():
            if key not in self.consumable:
                return None
            if self.consumable[key] < required.consumable[key]:
                return None

        return ResourcesInternal(
            {
                key: value - required.consumable.get(key, 0)
                for key, value in self.consumable.items()
            },
            self.non_consumable,
        )

    def add(self, returned: ResourcesInternal) -> ResourcesInternal:
        """
        Interpreting `self` as the available resources on an instance, adds back
        "resources required" for a job that has completed. returned.non_consumable is
        ignored.
        """
        return ResourcesInternal(
            {
                key: self.consumable.get(key, 0) + returned.consumable.get(key, 0)
                for key in set().union(self.consumable, returned.consumable)
            },
            self.non_consumable,
        )

    def combine(self, other: ResourcesInternal) -> ResourcesInternal:
        """
        combines two disjoint sets of resources. This is meant to combine instance type
        resources (like CPU and memory) with runtime resources (like AMI id and subnet
        id).

        I.e. the same resource cannot be present in self and other. This is different
        from add because add ignores non_consumable resources, and supports the case
        when self and other have the same consumable resources.
        """
        if other.consumable:
            new_consumable = self.consumable.copy()
            for key, value in other.consumable.items():
                if key in new_consumable:
                    raise ValueError(
                        f"Cannot combine resources, consumable resource {key} was in "
                        "both resources"
                    )
                new_consumable[key] = value
        else:
            new_consumable = self.consumable

        if other.non_consumable:
            new_non_consumable = self.non_consumable.copy()
            for key, value in other.non_consumable.items():
                if key in new_non_consumable:
                    raise ValueError(
                        f"Cannot combine resources, non-consumable resource {key} was "
                        "in both resources"
                    )
                new_non_consumable[key] = value
        else:
            new_non_consumable = self.non_consumable

        return ResourcesInternal(new_consumable, new_non_consumable)

    def divide_by(self, required: ResourcesInternal) -> int:
        """
        Interpreting `self` as available resources on an instance, returns how many
        tasks that require `required` resources could fit into this instance
        """
        # if a non_consumable required resource isn't available, then we won't be able
        # to fit any tasks
        for key, value in required.non_consumable.items():
            if key not in self.non_consumable:
                return 0
            if self.non_consumable[key] < required.non_consumable[key]:
                return 0

        # find the "binding" consumable that will determine how many tasks can fit on
        # this instance
        result = float(sys.maxsize)
        for key, value in required.consumable.items():
            if key not in self.consumable:
                return 0
            result = min(result, self.consumable[key] / required.consumable[key])

        return math.floor(result)

    def multiply(self, n: int) -> ResourcesInternal:
        """
        available.subtract(required.multiply(2)) should be equivalent to
        available.subtract(required).subtract(required)
        """
        return ResourcesInternal(
            {key: value * n for key, value in self.consumable.items()},
            self.non_consumable,
        )

    def non_standard_consumables(self) -> Iterable[Tuple[str, float]]:
        return (
            item
            for item in self.consumable.items()
            if item[0] != LOGICAL_CPU and item[0] != MEMORY_GB
        )

    def format_cpu_memory_gpu(self) -> str:
        if GPU in self.consumable:
            gpu_string = f", {self.consumable[GPU]} GPU"
        else:
            gpu_string = ""

        return (
            f"{self.consumable[LOGICAL_CPU]} CPU, {self.consumable[MEMORY_GB]} GB"
            + gpu_string
        )

    def format_eviction_rate(self) -> str:
        return f"{100 - self.non_consumable[EVICTION_RATE_INVERSE]}% eviction rate"

    def consumable_as_decimals(self) -> Dict[str, decimal.Decimal]:
        return {key: decimal.Decimal(value) for key, value in self.consumable.items()}

    def non_consumable_as_decimals(self) -> Dict[str, decimal.Decimal]:
        return {
            key: decimal.Decimal(value) for key, value in self.non_consumable.items()
        }

    def __str__(self) -> str:
        # TODO implement better
        other_consumables_str = ", ".join(
            f"{key}: {value}"
            for key, value in self.consumable.items()
            if key not in (LOGICAL_CPU, MEMORY_GB, GPU)
        )
        if other_consumables_str:
            other_consumables_str = ", " + other_consumables_str
        other_non_consumables_str = ", ".join(
            f"{key}: {value}"
            for key, value in self.non_consumable.items()
            if key != EVICTION_RATE_INVERSE
        )
        if other_non_consumables_str:
            other_non_consumables_str = ", " + other_non_consumables_str
        return (
            "Resources: consumables("
            f"{self.format_cpu_memory_gpu()}{other_consumables_str}), "
            f"non_consumables("
            f"{self.format_eviction_rate()}{other_non_consumables_str}"
            f")"
        )

    def copy(self) -> ResourcesInternal:
        return ResourcesInternal(self.consumable.copy(), self.non_consumable.copy())

    @classmethod
    def from_decimals(
        cls,
        consumable: Dict[str, decimal.Decimal],
        non_consumable: Dict[str, decimal.Decimal],
    ) -> ResourcesInternal:
        return cls(
            {key: float(value) for key, value in consumable.items()},
            {key: float(value) for key, value in non_consumable.items()},
        )

    @classmethod
    def from_cpu_and_memory(
        cls,
        logical_cpu: float,
        memory_gb: float,
        max_eviction_rate: Optional[float] = None,
        gpus: Optional[float] = None,
        gpu_memory: Optional[float] = None,
        flags_required: Union[Iterable[str], str, None] = None,
        *,
        other_consumables: Optional[Dict[str, float]] = None,
        other_non_consumables: Optional[Dict[str, float]] = None,
    ) -> ResourcesInternal:
        """A friendly constructor"""
        if other_consumables is None:
            consumables = {}
        else:
            consumables = other_consumables.copy()
        consumables[LOGICAL_CPU] = logical_cpu
        consumables[MEMORY_GB] = memory_gb
        if gpus is not None:
            consumables[GPU] = gpus
        if gpu_memory is not None:
            consumables[GPU_MEMORY] = gpu_memory

        if other_non_consumables is None:
            non_consumables = {}
        else:
            non_consumables = other_non_consumables.copy()

        if max_eviction_rate is not None:
            non_consumables[EVICTION_RATE_INVERSE] = 100.0 - max_eviction_rate

        if flags_required is not None:
            if isinstance(flags_required, str):
                flags_required = (flags_required,)
            for flag in flags_required:
                avx_version = avx_from_string(flag)
                if avx_version is not None:
                    non_consumables[AVX] = float(avx_version)
                else:
                    non_consumables[flag] = 1.0

        return cls(consumables, non_consumables)


def remaining_resources_sort_key(
    available_resources: ResourcesInternal, resources_required: ResourcesInternal
) -> Tuple[int, Optional[Tuple[float, float]]]:
    """
    This takes the available resources for an agent and returns (indicator,
    remaining_resources_sort_key).

    remaining_resources_sort_key indicates how "good" it is to run the job on that agent
    based on how many resources would be left if we created a worker for that job on
    that agent. The general idea is that having fewer resources left is better. E.g. if
    you have agent1 with 8GB+4cpu and agent2 with 2GB+1cpu, and you have a job that
    requires 2GB+1cpu, it's probably better to run it on agent2--that way if another job
    comes along that requires 8GB+4cpu, agent1 is still available to run that job. I.e.
    we prefer to create workers on agents that have a smaller
    remaining_resources_sort_key.

    indicator is either 0 or 1--0 means that it is possible to run the job on this
    agent, whereas 1 means that there are not enough resources to run the job on this
    agent. (We use the indicator this way because smaller is better for the
    remaining_resources_sort_key, so 0 is good and 1 is bad.)

    TODO we should probably be more thoughtful about the exact implementation of
     remaining_resources_sort_key.

    TODO it's possible we should "reserve" availability for future jobs that might come
     along. E.g. if we have many low-priority jobs that require 1 CPU, and then a
     high-priority job requiring 16 CPUs comes along we might wish that we had not used
     all of our resources to run the low-priority jobs. On the other hand, if you
     reserve resources and no future jobs come along, then you're just making the
     current job run slower for no reason.
    """
    remaining_resources = available_resources.subtract(resources_required)
    if remaining_resources is not None:
        # 0 is an indicator saying we can run this job
        return 0, (
            # TODO take non-consumables into account here?
            sum(value for key, value in remaining_resources.non_standard_consumables()),
            remaining_resources.consumable[MEMORY_GB]
            + 2 * remaining_resources.consumable[LOGICAL_CPU],
        )
    else:
        # 1 is an indicator saying we cannot run this job
        return 1, None


@dataclasses.dataclass(frozen=True)
class CloudInstanceType:
    name: str  # e.g. t2.micro
    on_demand_or_spot: OnDemandOrSpotType
    price: float  # 0.023 means 0.023 USD per hour to run the instance

    # Must include consumables:
    # - MEMORY_GB (e.g. 4 means 4 GiB)
    # - LOGICAL CPU (2 means 2 logical aka virtual cpus)
    # And non_consumables:
    # - EVICTION_RATE_INVERSE (e.g. 100 for on-demand instances, <100 for spot instances
    #   as (100 - percentage), so e.g. EVICTION_RATE_INVERSE==75 means 25% eviction rate
    #   aka probability of interruption)
    # TODO EVICTION_RATE_INVERSE should always use the latest data rather than always
    # using the number from when the instance was launched
    resources: ResourcesInternal

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "on_demand_or_spot": self.on_demand_or_spot,
            "price": self.price,
            "resources": {
                "consumable": self.resources.consumable,
                "non_consumable": self.resources.non_consumable,
            },
        }

    @classmethod
    def from_dict(cls, cit: Dict[str, Any]) -> CloudInstanceType:
        return cls(
            cit["name"],
            cit["on_demand_or_spot"],
            float(cit["price"]),
            ResourcesInternal(
                cit["resources"]["consumable"],
                cit["resources"]["non_consumable"],
            ),
        )


@dataclasses.dataclass
class ChosenCloudInstanceType:
    instance_type: CloudInstanceType
    # e.g. 5 means we should allocate 5 of these instances. Will always be >1
    num_instances: int

    # The maximum number of workers that you could put on this instance type
    workers_per_instance_full: int
    # price per worker if we fit as many workers as possible onto the instance
    price_per_worker_full: float

    # for "internal" use:
    workers_per_instance_current: int
    price_per_worker_current: float


@dataclasses.dataclass(frozen=True)
class CloudInstance:
    """Represents an instance launched by an InstanceRegistrar"""

    public_dns_name: str
    name: str

    instance_type: ChosenCloudInstanceType


def choose_instance_types_for_job(
    resources_required: ResourcesInternal,
    num_workers_to_allocate: int,
    original_instance_types: Iterable[CloudInstanceType],
) -> Iterable[ChosenCloudInstanceType]:
    """
    This chooses how many of which instance types we should launch for a job with 1 or
    more tasks where each task requires resources_required so that
    num_workers_to_allocate tasks can run in parallel. We choose the cheapest instances
    that have an eviction rate lower than or equal to the specified threshold. If you
    only want to use on-demand instances that have 0 eviction rate, you can set
    max_eviction_rate to 0. If there are multiple instances that are the cheapest, we
    choose the ones that have the lowest eviction rate. If there are still multiple
    instances, then we diversify among those instance types (it seems that evictions are
    more likely to happen at the same time on the same instance types).

    TODO we should maybe have an option where e.g. if you want to allocate 53 workers
     worth of capacity for a 100-task job, it makes more sense to allocate e.g. 55 or 60
     workers worth of capacity rather than allocating a little machine for the last 3
     workers of capacity
    """

    instance_types = []

    for orig_instance_type in original_instance_types:
        workers_per_instance_full = orig_instance_type.resources.divide_by(
            resources_required
        )
        # ignore instance types where we won't be able to fit even 1 worker
        if workers_per_instance_full >= 1:
            # if we get the maximum number of workers packed onto the instance type,
            # what is our effective price per worker
            price_per_worker_full = orig_instance_type.price / workers_per_instance_full

            instance_types.append(
                ChosenCloudInstanceType(
                    orig_instance_type,
                    0,
                    workers_per_instance_full,
                    price_per_worker_full,
                    workers_per_instance_full,
                    price_per_worker_full,
                )
            )

    # no instance types can run even one worker
    if len(instance_types) == 0:
        return []

    while num_workers_to_allocate > 0:
        # for larger instances, there might not be enough num_workers_to_allocate to
        # make it "worth it" to use that larger instance because we won't need enough
        # workers to fully pack the instance. So we recompute price_per_worker for those
        # instances assuming we only get to put num_workers_to_allocate on that
        # instance. We use ChosenCloudInstanceType.workers_per_instance_current and
        # price_per_worker_current to keep track of this.
        for instance_type in instance_types:
            if instance_type.workers_per_instance_full > num_workers_to_allocate:
                instance_type.price_per_worker_current = (
                    instance_type.instance_type.price / num_workers_to_allocate
                )
                instance_type.workers_per_instance_current = num_workers_to_allocate
            # as long as num_workers_to_allocate decreases on every iteration, we should
            # never need to "reset" price_per_worker_current_iteration or
            # workers_per_instance_current_iteration

        # Now find the instance types that have the lowest price per worker. If there
        # are multiple instance types that have the same price per worker (or are within
        # half a penny per hour), then take the ones that have the lowest eviction rate
        # (within 1%)
        # TODO maybe the rounding should be configurable?
        best_price_per_worker = min(
            instance_type.price_per_worker_current for instance_type in instance_types
        )
        best = [
            instance_type
            for instance_type in instance_types
            if instance_type.price_per_worker_current - best_price_per_worker < 0.005
        ]
        best_eviction_rate = max(
            instance_type.instance_type.resources.non_consumable[EVICTION_RATE_INVERSE]
            for instance_type in best
        )
        best = [
            instance_type
            for instance_type in best
            if best_eviction_rate
            - instance_type.instance_type.resources.non_consumable[
                EVICTION_RATE_INVERSE
            ]
            < 1
        ]

        # At this point, best is the set of instance types that are the cheapest and
        # least eviction-likely for our workload. Next, we'll make sure to take one, so
        # that we definitely make progress on allocating workers to instances. After
        # that, though, if we still have workers left to allocate, we want to allocate
        # them round-robin to the remaining instance types in best--the diversity in
        # instance types is good because instances of the same type are more likely to
        # get interrupted at the same time according to
        # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/spot-fleet-allocation-strategy.html

        # iterating from largest to smallest will make things easier
        best.sort(
            key=lambda instance_type: instance_type.workers_per_instance_full,
            reverse=True,
        )

        # take the first one no matter what
        best[0].num_instances += 1
        num_workers_to_allocate -= best[0].workers_per_instance_current

        # Now that we've decreased num_workers_to_allocate, we need to make sure
        # price_per_worker is still accurate (i.e. num_workers_to_allocate could have
        # fallen below workers_per_instance). If there are any instance types that where
        # price_per_worker is still accurate, then we'll continue to allocate to those
        # in a round-robin-ish fashion. Once there are no more instance types where
        # price_per_worker is still accurate, we'll go through the loop again and
        # recompute price_per_worker.
        i = 0
        while True:
            best = [
                instance_type
                for instance_type in best
                if instance_type.workers_per_instance_current <= num_workers_to_allocate
            ]
            if len(best) == 0:
                break
            # this is...very inexact because best is changing as we iterate, but the
            # idea is to walk through the options in best one by one
            i = (i + 1) % len(best)
            best[i].num_instances += 1
            num_workers_to_allocate -= best[i].workers_per_instance_current

    return (
        instance_type
        for instance_type in instance_types
        if instance_type.num_instances > 0
    )



from __future__ import annotations

import argparse
import dataclasses
import os
from typing import Optional, List, Iterable, Dict, Type, TYPE_CHECKING

from kubernetes_asyncio import client as kubernetes_client

from meadowrun.func_worker_storage_helper import (
    MEADOWRUN_STORAGE_USERNAME,
    MEADOWRUN_STORAGE_PASSWORD,
)
from meadowrun.gcp_integration.google_storage import get_google_storage_bucket
from meadowrun.k8s_integration.k8s import StorageBucketSpec, STORAGE_TYPE
from meadowrun.k8s_integration.k8s_core import get_kubernetes_secret
from meadowrun.storage_grid_job import (
    get_generic_username_password_bucket,
    GenericStorageBucket,
)

if TYPE_CHECKING:
    from meadowrun.abstract_storage_bucket import AbstractStorageBucket


@dataclasses.dataclass(frozen=True)
class GenericStorageBucketSpec(StorageBucketSpec):
    """
    Specifies a bucket in an object storage system. The object storage must be
    S3-compatible and use username/password authentication. The arguments provided will
    be used along the lines of:

    ```python
    import boto3
    boto3.Session(
        aws_access_key_id=username, aws_secret_access_key=password
    ).client(
        "s3", endpoint_url=endpoint_url
    ).download_file(
        Bucket=bucket, Key="test.file", Filename="test.file"
    )
    ```

    `username` and `password` should be the values provided by
    `username_password_secret`. (boto3 is built to be used with AWS S3, but it should
    work with any S3-compatible object store like Minio, Ceph, etc.)

    Attributes:
        bucket: The name of the bucket to use
        endpoint_url: The endpoint_url for the object storage system
        endpoint_url_in_cluster: Defaults to None which means use endpoint_url. You can
            set this to a different URL if you need to use a different URL from inside
            the Kubernetes cluster to access the storage endpoint
        username_password_secret: This should be the name of a Kubernetes secret
            that has a "username" and "password" key, where the username and password
            can be used to authenticate with the storage API.
    """

    bucket: str
    endpoint_url: str
    endpoint_url_in_cluster: Optional[str] = None
    username_password_secret: Optional[str] = None

    def _get_storage_endpoint_url_in_cluster(self) -> str:
        if self.endpoint_url_in_cluster is not None:
            return self.endpoint_url_in_cluster
        return self.endpoint_url

    async def get_storage_bucket(
        self, kubernetes_namespace: str
    ) -> AbstractStorageBucket:
        if self.username_password_secret is not None:
            secret_data = await get_kubernetes_secret(
                kubernetes_namespace,
                self.username_password_secret,
            )
        else:
            secret_data = {}

        return get_generic_username_password_bucket(
            self.endpoint_url,
            secret_data.get("username", None),
            secret_data.get("password", None),
            self.bucket,
        )

    def get_command_line_arguments(self) -> List[str]:
        return [
            "--storage-bucket",
            self.bucket,
            "--storage-endpoint-url",
            self._get_storage_endpoint_url_in_cluster(),
        ]

    def get_environment_variables(self) -> Iterable[kubernetes_client.V1EnvVar]:
        if self.username_password_secret is not None:
            yield kubernetes_client.V1EnvVar(
                name=MEADOWRUN_STORAGE_USERNAME,
                value_from=kubernetes_client.V1EnvVarSource(
                    secret_key_ref=kubernetes_client.V1SecretKeySelector(
                        key="username",
                        name=self.username_password_secret,
                        optional=False,
                    )
                ),
            )

            yield kubernetes_client.V1EnvVar(
                name=MEADOWRUN_STORAGE_PASSWORD,
                value_from=kubernetes_client.V1EnvVarSource(
                    secret_key_ref=kubernetes_client.V1SecretKeySelector(
                        key="password",
                        name=self.username_password_secret,
                        optional=False,
                    )
                ),
            )

    @classmethod
    def get_storage_type(cls) -> str:
        return "generic"

    @classmethod
    def add_arguments_to_parser(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--storage-bucket")
        parser.add_argument("--storage-endpoint-url")

    @classmethod
    async def from_parsed_args(cls, args: argparse.Namespace) -> GenericStorageBucket:
        # prepare storage client, filenames and pickle protocol for the result
        storage_username = os.environ.get(MEADOWRUN_STORAGE_USERNAME, None)
        storage_password = os.environ.get(MEADOWRUN_STORAGE_PASSWORD, None)
        if not args.storage_endpoint_url or not args.storage_bucket:
            raise ValueError(
                "--storage-endpoint-url and --storage-bucket must be specified with "
                f"{STORAGE_TYPE} {cls.get_storage_type()}"
            )
        return get_generic_username_password_bucket(
            args.storage_endpoint_url,
            storage_username,
            storage_password,
            args.storage_bucket,
        )


@dataclasses.dataclass(frozen=True)
class GoogleBucketSpec(StorageBucketSpec):
    """
    Specifies a bucket in Google Cloud Storage. Requires that credentials are available.
    This usually means that on the client you've logged in via the Google Cloud CLI, and
    in the Kubernetes cluster you're running with a service account (via
    the pod_customization parameter on Kubernetes) that has access to the specified
    bucket.

    Attributes:
        bucket: The name of the bucket to use
    """

    bucket: str

    async def get_storage_bucket(
        self, kubernetes_namespace: str
    ) -> AbstractStorageBucket:
        return get_google_storage_bucket(self.bucket)

    def get_command_line_arguments(self) -> List[str]:
        return ["--google-bucket", self.bucket]

    def get_environment_variables(self) -> Iterable[kubernetes_client.V1EnvVar]:
        return ()

    @classmethod
    def get_storage_type(cls) -> str:
        return "gcp"

    @classmethod
    def add_arguments_to_parser(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--google-bucket")

    @classmethod
    async def from_parsed_args(cls, args: argparse.Namespace) -> AbstractStorageBucket:
        return get_google_storage_bucket(args.google_bucket)


_ALL_STORAGE_BUCKET_SPEC_TYPES: List[Type[StorageBucketSpec]] = [
    GenericStorageBucketSpec,
    GoogleBucketSpec,
]
_ALL_STORAGE_BUCKET_SPECS: Dict[str, Type[StorageBucketSpec]] = {
    spec_type.get_storage_type(): spec_type
    for spec_type in _ALL_STORAGE_BUCKET_SPEC_TYPES
}


def add_storage_spec_arguments_to_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(STORAGE_TYPE, required=True)
    for spec_type in _ALL_STORAGE_BUCKET_SPECS.values():
        spec_type.add_arguments_to_parser(parser)


async def storage_bucket_from_parsed_args(
    args: argparse.Namespace,
) -> AbstractStorageBucket:
    spec_type = _ALL_STORAGE_BUCKET_SPECS[args.storage_type]
    return await spec_type.from_parsed_args(args)



"""Manages local deployments for run_job_local"""

from __future__ import annotations

import asyncio.subprocess
import functools
import hashlib
import os
import shutil
import tempfile
import urllib.parse
import zipfile
from typing import (
    Awaitable,
    Callable,
    List,
    Mapping,
    Optional,
    Sequence,
    TYPE_CHECKING,
    Tuple,
    Union,
)

import filelock

import meadowrun.func_worker_storage_helper
from meadowrun._vendor.aiodocker import exceptions as aiodocker_exceptions
from meadowrun.aws_integration.ecr import (
    get_ecr_helper,
)
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    _MEADOWRUN_GENERATED_DOCKER_REPO,
)
from meadowrun.azure_integration.acr import get_acr_helper
from meadowrun.credentials import RawCredentials, SshKey
from meadowrun.deployment.conda import get_cached_or_create_conda_environment
from meadowrun.deployment.pip import get_cached_or_create_pip_environment
from meadowrun.deployment.poetry import get_cached_or_create_poetry_environment
from meadowrun.deployment.prerequisites import (
    EnvironmentSpecPrerequisites,
    GOOGLE_AUTH_PACKAGE,
)
from meadowrun.docker_controller import (
    _does_digest_exist_locally,
    build_image,
    pull_image,
    push_image,
)
from meadowrun.meadowrun_pb2 import (
    CodeZipFile,
    EnvironmentSpec,
    EnvironmentSpecInCode,
    EnvironmentType,
    Job,
    ServerAvailableContainer,
    ServerAvailableInterpreter,
)
from meadowrun.run_job_core import (
    CloudProviderType,
    ContainerRegistryHelper,
)
from meadowrun.storage_grid_job import download_chunked_file
from meadowrun.abstract_storage_bucket import AbstractStorageBucket

if TYPE_CHECKING:
    StorageBucketFactoryType = Union[
        AbstractStorageBucket, Callable[[], Awaitable[AbstractStorageBucket]], None
    ]

_GIT_REPO_URL_SUFFIXES_TO_REMOVE = [".git", "/"]


async def _run_git(
    args: List[str], cwd: str, credentials: Optional[RawCredentials]
) -> Tuple[str, str]:
    """Runs a git command in an external process. Returns stdout, stderr"""
    env = os.environ.copy()

    if credentials is None:
        env["GIT_SSH_COMMAND"] = "ssh -o StrictHostKeyChecking=no"

        p = await asyncio.create_subprocess_exec(
            # TODO make the location of the git executable configurable
            "git",
            *args,
            cwd=cwd,
            env=env,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await p.communicate()
    elif isinstance(credentials, SshKey):
        fd, filename = tempfile.mkstemp()
        try:
            os.write(fd, credentials.private_key.encode("utf-8"))
            os.close(fd)

            # TODO investigate rules for GIT_SSH_COMMAND escaping on Windows and Linux
            escaped_private_key_file = filename.replace("\\", "\\\\")
            # TODO perhaps warn if GIT_SSH_COMMAND is somehow already populated?
            env["GIT_SSH_COMMAND"] = (
                f"ssh -i {escaped_private_key_file} -o IdentitiesOnly=yes "
                f"-o StrictHostKeyChecking=no"
            )

            p = await asyncio.create_subprocess_exec(
                "git",
                *args,
                cwd=cwd,
                env=env,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await p.communicate()
        finally:
            # TODO we should try even harder to delete the temporary file even if the
            #  process crashes before getting to this finally
            os.remove(filename)

    else:
        # TODO we should add support for username/password, e.g. from an OAuth flow
        raise ValueError(f"Unknown type of RawCredentials {type(credentials)}")

    if p.returncode != 0:
        raise ValueError(
            f"git {' '.join(args)} failed, return code {p.returncode}: "
            + stderr.decode()
        )

    # TODO lookup whether we should specify an encoding here?
    return stdout.decode(), stderr.decode()


async def get_code_paths(
    git_repos_folder: str,
    local_copies_folder: str,
    job: Job,
    credentials: Optional[RawCredentials],
    storage_bucket_factory: StorageBucketFactoryType,
) -> Tuple[Sequence[str], Optional[str], Optional[str]]:
    """
    Returns [code_paths, interpreter_spec_path, current_working_directory] based on
    Job.code_deployment. code_paths will have paths to folders that are available on
    this machine that contain "the user's code". interpreter_spec_path will be a path
    where we can find e.g. an environment.yml file that can be used to build a conda
    environment.

    For ServerAvailableFolder, this function will effectively use code_paths as
    specified.

    For GitRepoCommit, this function will create an immutable local copy of the specific
    commit.
    """
    case = job.WhichOneof("code_deployment")
    if case == "server_available_folder":
        if job.server_available_folder.code_paths:
            interpreter_spec_path = job.server_available_folder.code_paths[0]
        else:
            interpreter_spec_path = None
        return (
            job.server_available_folder.code_paths,
            # TODO theoretically the environment file could be not in the first code
            # path, but should be a very unlikely scenario for now.
            interpreter_spec_path,
            None,
        )
    elif case == "git_repo_commit":
        return await _get_git_code_paths(
            git_repos_folder,
            local_copies_folder,
            job.git_repo_commit.repo_url,
            job.git_repo_commit.commit,
            job.git_repo_commit.path_to_source,
            credentials,
        )
    elif case == "git_repo_branch":
        # warning this is not reproducible!!! should ideally be resolved on the
        # client
        return await _get_git_code_paths(
            git_repos_folder,
            local_copies_folder,
            job.git_repo_branch.repo_url,
            # TODO this is a bit of a hack. We don't want to bother pulling/merging e.g.
            # origin/main into main, so we just always reference origin/main.
            f"origin/{job.git_repo_branch.branch}",
            job.git_repo_branch.path_to_source,
            credentials,
        )
    elif case == "code_zip_file":
        return await _get_zip_file_code_paths(
            local_copies_folder, job.code_zip_file, storage_bucket_factory
        )
    else:
        raise ValueError(f"Unrecognized code_deployment {case}")


def _get_git_repo_local_clone_name(repo_url: str) -> str:
    """Gets a name to use for cloning the specified URL locally"""

    # try to normalize the url a little bit
    # TODO also consider .lower() in some cases?
    suffix_removed = True
    while suffix_removed:
        suffix_removed = False
        for s in _GIT_REPO_URL_SUFFIXES_TO_REMOVE:
            if repo_url.endswith(s):
                repo_url = repo_url[: -len(s)]
                suffix_removed = True

    return hashlib.blake2b(repo_url.encode("utf-8"), digest_size=16).hexdigest()


_DEFAULT_GIT_OPERATION_LOCK_TIMEOUT_SECS = 10 * 60


async def _get_git_code_paths(
    git_repos_folder: str,
    local_copies_folder: str,
    repo_url: str,
    revision_spec: str,
    path_to_source: str,
    credentials: Optional[RawCredentials],
) -> Tuple[Sequence[str], str, str]:
    """Returns code_paths for GitRepoCommit"""

    local_clone_name = _get_git_repo_local_clone_name(repo_url)
    with filelock.FileLock(
        os.path.join(git_repos_folder, f"{local_clone_name}.lock"),
        _DEFAULT_GIT_OPERATION_LOCK_TIMEOUT_SECS,
    ):
        # clone the repo locally/update it
        local_path = os.path.join(git_repos_folder, local_clone_name)

        if os.path.exists(local_path):
            _ = await _run_git(["fetch"], local_path, credentials)
        else:
            # TODO do something with output?
            temp_path = f"{local_path}_{os.getpid()}"
            _ = await _run_git(
                ["clone", repo_url, temp_path], git_repos_folder, credentials
            )
            os.rename(temp_path, local_path)

        # TODO we should (maybe) prevent very ambiguous specifications like HEAD
        out, err = await _run_git(["rev-parse", revision_spec], local_path, credentials)
        commit_hash = out.strip()
        # TODO do something with output?
        # get and checkout the specified hash
        _ = await _run_git(["checkout", commit_hash], local_path, credentials)

        # copy the specified version of this repo from local_path into
        # local_copy_path

        local_copy_path = os.path.join(
            local_copies_folder, local_clone_name + "_" + commit_hash
        )

        # it's important that it's impossible to create a scenario where different
        # local_clone.local_name + commit_hash can result in identical strings
        if not os.path.exists(local_copy_path):
            # TODO really we should do the whole thing where we hash each file/folder
            # and create symlinks so that we don't end up with tons of copies of
            # identical files
            shutil.copytree(local_path, local_copy_path)

        # TODO raise a friendlier exception if this path doesn't exist
        source_path = os.path.realpath((os.path.join(local_copy_path, path_to_source)))
        return ([source_path], local_copy_path, local_copy_path)


async def _get_zip_file_code_paths(
    local_copies_folder: str,
    code_zip_file: CodeZipFile,
    storage_bucket_factory: StorageBucketFactoryType,
) -> Tuple[List[str], None, str]:
    """
    code_zip_file.url will be the URL of a file in the object storage system as
    generated by _upload_code_zip_file. This function should download the file and
    extract it to local_copies_folder if it has not already been extracted.
    """
    decoded_url = urllib.parse.urlparse(code_zip_file.url)
    if decoded_url.scheme != "mdrstorage":  # just a sanity check
        raise ValueError(f"Unexpected URL scheme in {code_zip_file.url}")

    object_name = decoded_url.path.lstrip("/")
    extracted_folder = os.path.join(local_copies_folder, os.path.basename(object_name))

    with filelock.FileLock(f"{extracted_folder}.lock", timeout=120):
        if not os.path.exists(extracted_folder):
            zip_file_path = extracted_folder + ".zip"
            if storage_bucket_factory is None:
                raise ValueError(
                    "storage_bucket_factory must be set to download CodeZipFiles"
                )
            elif isinstance(storage_bucket_factory, AbstractStorageBucket):
                await download_chunked_file(
                    storage_bucket_factory, object_name, zip_file_path
                )
            else:
                async with await storage_bucket_factory() as storage_bucket:
                    await download_chunked_file(
                        storage_bucket, object_name, zip_file_path
                    )
            with zipfile.ZipFile(zip_file_path) as zip_file:
                zip_file.extractall(extracted_folder)

    return (
        [
            os.path.join(local_copies_folder, extracted_folder, zip_path)
            for zip_path in code_zip_file.code_paths
        ],
        None,
        os.path.join(local_copies_folder, extracted_folder, code_zip_file.cwd_path),
    )


async def compile_environment_spec_to_container(
    environment_spec: Union[EnvironmentSpecInCode, EnvironmentSpec],
    interpreter_spec_path: str,
    cloud: Optional[Tuple[CloudProviderType, str]],
) -> ServerAvailableContainer:
    """
    Turns e.g. a conda_environment.yml file into a docker container which will be
    available locally (as per the returned ServerAvailableContainer), and also cached in
    ECR.

    TODO we should also consider creating conda environments locally rather than always
    making a container for them, that might be more efficient.
    """
    path_to_spec, spec_hash, prerequisites = _get_path_and_hash(
        environment_spec, interpreter_spec_path
    )

    if cloud is None:
        helper = ContainerRegistryHelper(
            False, None, f"{_MEADOWRUN_GENERATED_DOCKER_REPO}:{spec_hash}", False
        )
    else:
        cloud_type, region_name = cloud
        if cloud_type == "EC2":
            helper = await get_ecr_helper(
                _MEADOWRUN_GENERATED_DOCKER_REPO, spec_hash, region_name
            )
        elif cloud_type == "AzureVM":
            helper = await get_acr_helper(
                _MEADOWRUN_GENERATED_DOCKER_REPO, spec_hash, region_name
            )
        else:
            raise ValueError(f"Unexpected cloud_type {cloud_type}")

    result = ServerAvailableContainer(image_name=helper.image_name)

    # if the image already exists locally, just return it
    if await _does_digest_exist_locally(helper.image_name):
        return result

    # if the image doesn't exist locally but does exist in ECR, try to pull it
    if helper.does_image_exist:
        try:
            await pull_image(helper.image_name, helper.username_password)
            # TODO we're assuming that once we pull the image it will be available
            # locally until the job runs, which might not be true if we implement
            # something to clean unused images.
            return result
        except (aiodocker_exceptions.DockerError, ValueError) as e:
            print(
                f"Warning, image {helper.image_name} was supposed to exist, but we were"
                " unable to pull, so re-building locally. This could happen if the "
                "image was deleted between when we checked for it and when we pulled, "
                f"but that should be rare: {e}"
            )

    # the image doesn't exist locally or in ECR, so build it ourselves and cache it in
    # ECR

    print(f"Building python environment in container {spec_hash}")

    build_args = {}
    files_to_copy = []
    apt_packages = [
        p for p in environment_spec.additional_software.keys() if p != "cuda"
    ]
    if prerequisites & EnvironmentSpecPrerequisites.GIT:
        apt_packages.append("git")
    if apt_packages:
        build_args["APT_PACKAGES"] = " ".join(apt_packages)
    if prerequisites & EnvironmentSpecPrerequisites.GOOGLE_AUTH:
        build_args["ENVIRONMENT_PREREQUISITES"] = GOOGLE_AUTH_PACKAGE

    if environment_spec.environment_type == EnvironmentType.CONDA:
        docker_file_name = "CondaDockerfile"
        spec_filename = os.path.basename(path_to_spec)
        build_args["ENV_FILE"] = spec_filename
        files_to_copy.append((path_to_spec, spec_filename))
        if "cuda" in environment_spec.additional_software:
            build_args[
                "CONDA_IMAGE"
            ] = "meadowrun/cuda-conda:conda4.12.0-cuda11.6.2-cudnn8-ubuntu20.04"
        else:
            build_args["CONDA_IMAGE"] = "continuumio/miniconda3"
    else:
        # common to pip and poetry
        if "cuda" in environment_spec.additional_software:
            build_args["PYTHON_IMAGE"] = (
                f"meadowrun/cuda-python:py{environment_spec.python_version}-"
                f"cuda11.6.2-cudnn8-ubuntu20.04"
            )
        else:
            build_args[
                "PYTHON_IMAGE"
            ] = f"python:{environment_spec.python_version}-slim-bullseye"

        if environment_spec.environment_type == EnvironmentType.PIP:
            docker_file_name = "PipDockerfile"
            spec_filename = os.path.basename(path_to_spec)
            files_to_copy.append((path_to_spec, spec_filename))
            build_args["ENV_FILE"] = spec_filename
        elif environment_spec.environment_type == EnvironmentType.POETRY:
            docker_file_name = "PoetryDockerfile"
            files_to_copy.append(
                (os.path.join(path_to_spec, "pyproject.toml"), "pyproject.toml")
            )
            files_to_copy.append(
                (os.path.join(path_to_spec, "poetry.lock"), "poetry.lock")
            )
        else:
            raise ValueError(
                f"Unexpected environment_type {environment_spec.environment_type}"
            )

    docker_file_path = os.path.join(
        os.path.dirname(__file__), "docker_files", docker_file_name
    )
    files_to_copy.append((docker_file_path, "Dockerfile"))
    await build_image(files_to_copy, helper.image_name, build_args)

    # try to push the image so that we can reuse it later
    # TODO this should really happen asynchronously as it takes a long time and isn't
    # critical.
    # TODO we should also somehow avoid multiple processes on the same (or even on
    # different machines) building the identical image at the same time.
    if helper.should_push:
        try:
            print("Caching container")
            await push_image(helper.image_name, helper.username_password)
        except aiodocker_exceptions.DockerError as e:
            print(
                f"Warning, image {helper.image_name} couldn't be pushed. Execution can "
                f"continue, but we won't be able to reuse this image later: {e}"
            )

    return result


async def compile_environment_spec_locally(
    environment_spec: Union[EnvironmentSpecInCode, EnvironmentSpec],
    interpreter_spec_path: str,
    built_interpreters_folder: str,
) -> ServerAvailableInterpreter:
    """
    Turns e.g. a conda_environment.yml file into a locally available python interpreter
    """
    path_to_spec, spec_hash, prerequisites = _get_path_and_hash(
        environment_spec, interpreter_spec_path
    )
    # we don't have to worry about has_git_dependency as we assume we always have git
    # installed in this environment

    print(f"Building python environment locally {spec_hash}")

    # TODO we shouldn't install apt packages but we should check to see whether they're
    # installed
    # apt_packages = [
    #     p for p in environment_spec.additional_software.keys() if p != "cuda"
    # ]

    if environment_spec.environment_type == EnvironmentType.CONDA:
        get_cached_or_create: Callable[
            [
                str,
                str,
                EnvironmentSpecPrerequisites,
                str,
                Callable[[str, str], Awaitable[bool]],
                Callable[[str, str], Awaitable[None]],
            ],
            Awaitable[str],
        ] = get_cached_or_create_conda_environment
    elif environment_spec.environment_type == EnvironmentType.PIP:
        get_cached_or_create = get_cached_or_create_pip_environment
    elif environment_spec.environment_type == EnvironmentType.POETRY:
        get_cached_or_create = get_cached_or_create_poetry_environment
    else:
        raise ValueError(
            f"Unexpected environment_type {environment_spec.environment_type}"
        )

    # TODO this isn't the right way to do this--the storage client should be
    # getting passed in from higher up. For now this works because this code
    # path is only being hit from Kubernetes
    storage_bucket = meadowrun.func_worker_storage_helper.FUNC_WORKER_STORAGE_BUCKET
    if storage_bucket is None:
        raise ValueError("FUNC_WORKER_STORAGE_BUCKET was not set")
    return ServerAvailableInterpreter(
        interpreter_path=await get_cached_or_create(
            spec_hash,
            path_to_spec,
            prerequisites,
            os.path.join(built_interpreters_folder, spec_hash),
            # try_get_file(remote_file_name: str, local_file_name: str) -> bool. Tries
            # to download the specified remote file to the specified local file name.
            # Returns True if the file is available, False if the file is not available.
            functools.partial(storage_bucket.try_get_file),
            # upload_file(local_file_name: str, remote_file_name: str) -> None. Uploads
            # the specified file, overwrites any existing remote file.
            functools.partial(storage_bucket.write_file),
        )
    )


def _hash_spec(spec_contents: bytes) -> str:
    # TODO better to exclude the name and prefix for conda yml files as those
    # are ignored
    return hashlib.blake2b(spec_contents, digest_size=64).hexdigest()


def _format_additional_software(additional_software: Mapping[str, str]) -> str:
    return "\n\nmeadowrun-additional-software: " + "; ".join(
        f"{key}: {value}" for key, value in sorted(additional_software.items())
    )


def _get_path_and_hash(
    environment_spec: Union[EnvironmentSpecInCode, EnvironmentSpec],
    interpreter_spec_path: str,
) -> Tuple[str, str, EnvironmentSpecPrerequisites]:
    """
    Returns path_to_spec, spec_hash, has_git_dependency. spec_hash is a hash that
    uniquely identifies the environment. path_to_spec is used slightly differently for
    each environment type's Dockerfile. has_git_dependency is also specific to each
    """

    if isinstance(environment_spec, EnvironmentSpecInCode):
        # in this case, interpreter_spec_path is a path to the root of the place where
        # the spec lives.
        path_to_spec = os.path.join(
            interpreter_spec_path, environment_spec.path_to_spec
        )
        if environment_spec.environment_type == EnvironmentType.POETRY:
            # in the case of Poetry, the path_to_spec should be a folder that contains
            # pyproject.toml and poetry.lock
            file_to_hash = os.path.join(path_to_spec, "poetry.lock")
        else:
            file_to_hash = path_to_spec
        with open(file_to_hash, "rb") as spec:
            spec_contents_bytes = spec.read()
        return (
            path_to_spec,
            _hash_spec(
                spec_contents_bytes
                + _format_additional_software(
                    environment_spec.additional_software
                ).encode("utf-8")
            ),
            _get_environment_spec_prerequisites(
                spec_contents_bytes.decode("utf-8"), environment_spec.environment_type
            ),
        )
    elif isinstance(environment_spec, EnvironmentSpec):
        # in this case, interpreter_spec_path is a path to where we save the spec for
        # later processing.
        if environment_spec.environment_type == EnvironmentType.POETRY:
            # for poetry, path_to_spec will be a folder that contains pyproject.toml and
            # poetry.lock
            spec_contents_str = environment_spec.spec_lock
            spec_hash = _hash_spec(
                (
                    spec_contents_str
                    + _format_additional_software(environment_spec.additional_software)
                ).encode("UTF-8")
            )
            path_to_spec = os.path.join(interpreter_spec_path, spec_hash)
            os.makedirs(path_to_spec, exist_ok=True)
            with open(
                os.path.join(path_to_spec, "pyproject.toml"), "w", encoding="utf-8"
            ) as spec_file:
                spec_file.write(environment_spec.spec)
            with open(
                os.path.join(path_to_spec, "poetry.lock"), "w", encoding="utf-8"
            ) as lock_file:
                lock_file.write(environment_spec.spec_lock)
        else:
            spec_contents_str = environment_spec.spec
            spec_hash = _hash_spec(
                (
                    spec_contents_str
                    + _format_additional_software(environment_spec.additional_software)
                ).encode("UTF-8")
            )

            if environment_spec.environment_type == EnvironmentType.CONDA:
                spec_filename = f"conda_env_spec_{spec_hash}.yml"
            elif environment_spec.environment_type == EnvironmentType.PIP:
                spec_filename = f"pip_env_spec_{spec_hash}.txt"
            else:
                raise ValueError(
                    f"Unexpected environment_type: {environment_spec.environment_type}"
                )

            path_to_spec = os.path.join(interpreter_spec_path, spec_filename)
            with open(path_to_spec, "w", encoding="utf-8") as env_spec:
                env_spec.write(environment_spec.spec)

        return (
            path_to_spec,
            spec_hash,
            _get_environment_spec_prerequisites(
                spec_contents_str, environment_spec.environment_type
            ),
        )
    else:
        raise ValueError(
            f"Unexpected type of environment_spec {type(environment_spec)}"
        )


def _get_environment_spec_prerequisites(
    spec_contents: str, environment_type: EnvironmentType.ValueType
) -> EnvironmentSpecPrerequisites:
    result = EnvironmentSpecPrerequisites.NONE

    if (
        environment_type == EnvironmentType.PIP
        or environment_type == EnvironmentType.CONDA
    ):
        # example line resulting from pip freeze:
        # package @ git+https://github.com/name/repo.git@sha
        # Conda can have git-based dependencies via pip
        if "git+" in spec_contents:
            result |= EnvironmentSpecPrerequisites.GIT

        if "-python.pkg.dev" in spec_contents:
            # TODO this could be more sophisticated, e.g. if they've already supplied a
            # username/password, if they're not running on GCP, then we shouldn't set
            # this flag
            result |= EnvironmentSpecPrerequisites.GOOGLE_AUTH
    elif environment_type == EnvironmentType.POETRY:
        for line in spec_contents.splitlines():
            if line.startswith('type = "git"'):
                result |= EnvironmentSpecPrerequisites.GIT
            elif line.startswith("url = ") and "-python.pkg.dev" in line:
                result |= EnvironmentSpecPrerequisites.GOOGLE_AUTH
            # TODO break early if we've found everything?

    return result



"""
Runs a single job locally. Meant to be used on a "server" where the client is calling
e.g. run_function


A long note about interrupting running jobs and deallocation, ssh, asyncio, signal
handlers, atexit handlers and cleanup.

When a job is running on a machine, via this script, the client may get interrupted or
otherwise exit. In that case, we'd like this script to exit as soon as possible, as it's
now using resources unnecessarily. Additionally, just exiting the process is not enough
- the job must also be deallocated: the table that has an overview of all running jobs
needs to be updated.

This script is always started via an ssh command, see run_job_core.py. There are now
roughly three possibilities:

1. The job runs to completion normally. In this case this script starts a process to
   deallocate, without waiting for it to finish. This way, users don't have to wait for
   deallocation but it happens asynchronously from their perspective.

2. The user interrupts the job from the client side. In this case we propagate the
   interrupt signal via ssh to server process, which reacts to it by cancelling whatever
   it is currently doing and finally starting the deallocation process.

3. The user kills the job from the client side. The difference with 2 is that in this
   case the client doesn't get an opportunity to send the interrupt signal to the server
   (this script). In this case sshd will eventually realize there's no ssh client
   connected to the session, and send SIGHUP to the shell session, which will eventually
   terminate the server process. To deal with such unexpected ends, each machine runs a
   cron job that deallocates processes that are no longer running perdiodically.

Some details about point 2:

- To be able to propagate the interruption from client to server, we obviously need to
  be able to react to a keyboard interrupt or cancellation.

- Interrupting an asyncio process is trickier than interrupting a non-asyncio process,
  because in the single-threaded, synchronous case, Python raises a
  KeyboardInterruptError which you can react to in an exception handler. In the asyncio
  case this error is still raised, but that doesn't necessarily mean that any suspended
  tasks get a chance to run. The program just exits with a bunch of unfinished tasks on
  the event loop. Since we use asyncssh, see ssh.py, we connect to remote hosts in an
  async coroutine. On a KeyboardInterrupt, we need those tasks to run so the
  asyncssh.Process can be sent a SIGINT in turn. Luckily, if you run your event loop
  through asyncio.run, this takes care of cancelling outstanding tasks and waiting for
  them automatically, so we can react to CancelledError and propagate the SIGINT. See
  ssh.py::run_and_print.

- Also, the asyncio coroutines should not swallow CancelledError (as some used to in
  run_job_local.py), as that means any coroutines that contain them do not get
  cancelled.

- On the server side, the SIGINT again gets interpreted as a KeyboardInterruptError, so
  the same mechanism via asyncio.run applies: all outstanding tasks get cancelled and
  get a chance to run. We take advantage of this mechanism below by starting the
  deallocate process in a finally block from the main task.

- Note that none of this works if either client or server gets SIGTERM (terminate) or
  SIGKILL (kill -9) signals - tasks will not be cleanly cancelled and run to completion,
  and the deallocate process will not run.

Some details about point 3:

- If a client abruptly disappears, it'd be better for the process to stop running and
  get deallocated as soon as possible. This will indeed eventually happen, but is
  unlikely to happen straight away for typical distributed systems reasons (i.e. you
  don't want processes exiting because of a temporary drop in connectivity). Here's a
  good link about the process:
  https://serverfault.com/questions/463366/does-getting-disconnected-from-an-ssh-session-kill-your-programs
  Via testing I found that outputting something to stdout is the sure way to get a
  "orphaned" ssh process to exit. This is not implemented, but we could print an
  invisible character (or a spinning ascii wheel) periodically as a poor man's
  keepalive. In testing, I've not been able to get sshd's options for keepalive checking
  to work:
  https://unix.stackexchange.com/questions/3026/what-do-options-serveraliveinterval-and-clientaliveinterval-in-sshd-config-d
  That said, it's entirely possible I made some basic mistake.

"""

from __future__ import annotations

import argparse
import asyncio
import functools
import logging
import os
import sys
import traceback
from typing import Optional, Tuple, TYPE_CHECKING

import meadowrun.run_job_local
from meadowrun.azure_integration.blob_storage import get_azure_blob_container
from meadowrun.meadowrun_pb2 import ProcessState, Job
from meadowrun.run_job_core import CloudProvider, CloudProviderType
from meadowrun.storage_grid_job import get_aws_s3_bucket_async
import meadowrun.func_worker_storage_helper

if TYPE_CHECKING:
    from meadowrun.deployment_manager import StorageBucketFactoryType


async def main_async(
    job_id: str,
    cloud: Optional[Tuple[CloudProviderType, str]],
) -> None:
    job_io_prefix = f"/var/meadowrun/io/{job_id}"
    try:
        # write to a temp file and then rename to make sure deallocate_tasks doesn't see
        # a partial write
        with open(f"{job_io_prefix}.pid_temp", mode="w", encoding="utf-8") as f:
            f.write(str(os.getpid()))
        os.rename(f"{job_io_prefix}.pid_temp", f"{job_io_prefix}.pid")

        with open(f"{job_io_prefix}.job_to_run", mode="rb") as f:
            bytes_job_to_run = f.read()
        job = Job()
        job.ParseFromString(bytes_job_to_run)

        if cloud is None:
            storage_bucket_factory: StorageBucketFactoryType = None
        elif cloud[0] == "EC2":
            storage_bucket_factory = functools.partial(
                get_aws_s3_bucket_async, cloud[1]
            )
        elif cloud[0] == "AzureVM":
            storage_bucket_factory = functools.partial(
                get_azure_blob_container, cloud[1]
            )
        else:
            print(f"Warning: unknown value for cloud {cloud}")
            storage_bucket_factory = None

        first_state, continuation = await meadowrun.run_job_local.run_local(
            job, cloud, storage_bucket_factory, compile_environment_in_container=True
        )
        with open(f"{job_io_prefix}.initial_process_state", mode="wb") as f:
            f.write(first_state.SerializeToString())

        if (
            first_state.state != ProcessState.ProcessStateEnum.RUNNING
            or continuation is None
        ):
            with open(f"{job_io_prefix}.process_state", mode="wb") as f:
                f.write(first_state.SerializeToString())
        else:
            final_process_state = await continuation
            # if the result is large it's a little sad because we're duplicating it into
            # this .process_state file
            with open(f"{job_io_prefix}.process_state", mode="wb") as f:
                f.write(final_process_state.SerializeToString())
    except (KeyboardInterrupt, asyncio.CancelledError):
        print("Job was killed by SIGINT")
    except:  # noqa: E722
        # so we know what's wrong
        traceback.print_exc()
        raise
    finally:
        if cloud is not None:
            # we want to kick this off and then allow the current process to complete
            # without affecting the child process. This seems to work on Linux but not
            # on Windows (but we don't currently support Windows, so that's okay)
            await asyncio.subprocess.create_subprocess_exec(
                sys.executable,
                os.path.join(os.path.dirname(__file__), "deallocate_jobs.py"),
                "--cloud",
                cloud[0],  # e.g. EC2 or Azure
                "--cloud-region-name",
                cloud[1],
                "--job-id",
                job_id,
            )


def main(
    job_id: str,
    cloud: Optional[Tuple[CloudProviderType, str]],
) -> None:
    try:
        asyncio.run(main_async(job_id, cloud))
    except (KeyboardInterrupt, asyncio.CancelledError):
        print("Job was killed by SIGINT")


def command_line_main() -> None:
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("--job-id", required=True)
    parser.add_argument("--cloud", choices=CloudProvider)
    parser.add_argument("--cloud-region-name")
    args = parser.parse_args()

    if bool(args.cloud is None) ^ bool(args.cloud_region_name is None):
        raise ValueError(
            "--cloud and --cloud-region-name must both be provided or both not be "
            "provided"
        )

    if args.cloud is None:
        cloud = None
    else:
        cloud = args.cloud, args.cloud_region_name

    main(args.job_id, cloud)


if __name__ == "__main__":
    command_line_main()



"""Specifies deployments for jobs"""

from __future__ import annotations

from typing import Union, cast, overload
from typing_extensions import Final, get_args

from meadowrun.credentials import (
    CredentialsDict,
    RawCredentials,
    get_docker_credentials,
)
from meadowrun.docker_controller import get_latest_digest_from_registry
from meadowrun.meadowrun_pb2 import (
    CodeZipFile,
    ContainerAtDigest,
    ContainerAtTag,
    EnvironmentSpec,
    EnvironmentSpecInCode,
    GitRepoBranch,
    GitRepoCommit,
    ServerAvailableContainer,
)
from meadowrun.meadowrun_pb2 import ServerAvailableFolder, ServerAvailableInterpreter

CodeDeployment = Union[ServerAvailableFolder, GitRepoCommit, CodeZipFile]
CodeDeploymentTypes: Final = get_args(CodeDeployment)


InterpreterDeployment = Union[
    ServerAvailableInterpreter,
    ContainerAtDigest,
    ServerAvailableContainer,
    # TODO this is kind of a VersionedInterpreterDeployment, but it can't be resolved
    # without instantiating the CodeDeployment. The whole
    # InterpreterDeployment/VersionedInterpreterDeployment distinction probably needs
    # rethinking.
    EnvironmentSpecInCode,
    EnvironmentSpec,
]
InterpreterDeploymentTypes: Final = get_args(InterpreterDeployment)


# Similar to a CodeDeployment, but instead of a single version of the code (e.g. a
# specific commit in a git repo), specifies a versioned codebase (e.g. a git repo) where
# we can e.g. get the latest, select an old version.
VersionedCodeDeployment = GitRepoBranch
VersionedCodeDeploymentTypes: Final = (VersionedCodeDeployment,)


# Similar to InterpreterDeployment, but instead of a single version of the interpreter
# (e.g. a specific digest in a container repository), specifies a versioned set of
# interpreters (e.g. an entire container repository) where we can e.g. get the latest or
# select an old version.
VersionedInterpreterDeployment = ContainerAtTag
VersionedInterpreterDeploymentTypes: Final = (VersionedCodeDeployment,)


# TODO the below methods are not currently used but they should be used by run_map so
# that we resolve to a consistent deployment before we start a grid job. E.g. you don't
# want the same ContainerAtTag resolving to different ContainerAtDigest for different
# workers.


@overload
async def get_latest_code_version(
    code: CodeDeployment,
    credentials: Union[CredentialsDict, RawCredentials, None],
) -> CodeDeployment:
    ...


@overload
async def get_latest_code_version(
    code: VersionedCodeDeployment,
    credentials: Union[CredentialsDict, RawCredentials, None],
) -> GitRepoCommit:
    ...


async def get_latest_code_version(
    code: Union[CodeDeployment, VersionedCodeDeployment],
    credentials: Union[CredentialsDict, RawCredentials, None],
) -> CodeDeployment:
    if isinstance(code, CodeDeploymentTypes):
        # mypy wouldn't need a cast if we added the types explicitly instead of
        # CodeDeploymentTypes...but then we'd need to repeat that.
        return cast(CodeDeployment, code)
    elif isinstance(code, GitRepoBranch):
        # TODO this is a stub for now. We should remove logic for translating a branch
        #  into a commit in _get_git_repo_commit_interpreter_and_code and implement it
        #  here using git ls-remote
        # TODO also we will add the ability to specify overrides (e.g. a specific commit
        #  or an alternate branch)
        return GitRepoCommit(
            repo_url=code.repo_url,
            commit="origin/" + code.branch,
            path_to_source=code.path_to_source,
        )
    else:
        raise ValueError(
            f"Unknown CodeDeployment/VersionedCodeDeployment type {type(code)}"
        )


@overload
async def get_latest_interpreter_version(
    interpreter: InterpreterDeployment,
    credentials: Union[CredentialsDict, RawCredentials, None],
) -> InterpreterDeployment:
    ...


@overload
async def get_latest_interpreter_version(
    interpreter: VersionedInterpreterDeployment,
    credentials: Union[CredentialsDict, RawCredentials, None],
) -> ContainerAtDigest:
    ...


async def get_latest_interpreter_version(
    interpreter: Union[InterpreterDeployment, VersionedInterpreterDeployment],
    credentials: Union[CredentialsDict, RawCredentials, None],
) -> InterpreterDeployment:
    if isinstance(interpreter, InterpreterDeploymentTypes):
        return cast(InterpreterDeployment, interpreter)
    elif isinstance(interpreter, ContainerAtTag):
        if credentials is not None and not isinstance(credentials, RawCredentials):
            credentials = await get_docker_credentials(
                interpreter.repository, credentials
            )
        return ContainerAtDigest(
            repository=interpreter.repository,
            digest=await get_latest_digest_from_registry(
                interpreter.repository,
                interpreter.tag,
                credentials,
            ),
        )
    else:
        raise ValueError(
            "Unknown InterpreterDeployment/VersionedInterpreterDeployment type "
            f"{type(interpreter)}"
        )



import base64
import re
from typing import Optional

import boto3

from meadowrun.aws_integration.aws_core import (
    _get_account_number,
    _get_default_region_name,
)
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    ignore_boto3_error_code,
)
from meadowrun.credentials import UsernamePassword
from meadowrun.run_job_core import ContainerRegistryHelper


def _get_username_password(region_name: str) -> UsernamePassword:
    """
    Returns (username, password) for the ECR default repository. Can be passed to
    functions in docker_controller.py
    """
    client = boto3.client("ecr", region_name=region_name)
    response = client.get_authorization_token()
    username_password = base64.b64decode(
        response["authorizationData"][0]["authorizationToken"]
    ).decode("utf-8")
    username, sep, password = username_password.partition(":")
    if not sep or not password:
        raise ValueError(
            "username_password was not in expected format username:password"
        )

    return UsernamePassword(username, password)


def _ensure_repository(repository: str, region_name: str) -> None:
    """Creates the ECR repository if it doesn't exist"""
    client = boto3.client("ecr", region_name=region_name)
    ignore_boto3_error_code(
        lambda: client.create_repository(repositoryName=repository),
        "RepositoryAlreadyExistsException",
    )


def _does_image_exist(repository: str, tag: str, region_name: str) -> bool:
    """
    Returns whether the specified image exists.

    It feels like you should be able to just see whether docker_controller:pull_image is
    successful, but that takes significantly longer, and you also don't know if a
    failure is the result of something else like an authentication issue.
    """
    client = boto3.client("ecr", region_name=region_name)
    success, result = ignore_boto3_error_code(
        lambda: client.describe_images(
            repositoryName=repository, imageIds=[{"imageTag": tag}]
        ),
        "ImageNotFoundException",
    )
    return success


async def get_ecr_helper(
    repository: str, tag: str, region_name: str
) -> ContainerRegistryHelper:
    """This function is tightly coupled with compile_environment_spec_to_container"""
    if region_name == "default":
        region_name = await _get_default_region_name()

    repository_prefix = f"{_get_account_number()}.dkr.ecr.{region_name}.amazonaws.com"

    return ContainerRegistryHelper(
        True,
        _get_username_password(region_name),
        f"{repository_prefix}/{repository}:{tag}",
        _does_image_exist(repository, tag, region_name),
    )


def get_ecr_username_password(registry_domain: str) -> Optional[UsernamePassword]:
    """Gets the username/password if registry_domain is an ECR repo"""

    match = re.match(
        r"\d+\.dkr\.ecr\.(?P<region_name>[^.]+)\.amazonaws\.com",
        registry_domain.lower(),
    )
    if match is None:
        return None

    region_name = match.groups("region_name")[0]
    return _get_username_password(region_name)



from __future__ import annotations

import asyncio
import json
import os
from typing import List, Optional, Tuple, Callable, Awaitable, TYPE_CHECKING

import filelock

from meadowrun.shared import remove_corrupted_environment
from meadowrun.storage_keys import STORAGE_ENV_CACHE_PREFIX

if TYPE_CHECKING:
    from meadowrun.deployment.prerequisites import EnvironmentSpecPrerequisites


class CondaMissingException(Exception):
    """Raised when the conda executable can not be found."""

    pass


async def _run(args: List[str]) -> Tuple[str, str]:
    """Runs a conda command in an external process. Returns stdout, stderr."""

    if "CONDA_EXE" not in os.environ:
        raise CondaMissingException(
            "CONDA_EXE not set - is conda installed and configured correctly?"
        )

    conda = os.environ["CONDA_EXE"]
    env = os.environ.copy()

    p = await asyncio.create_subprocess_exec(
        conda,
        *args,
        env=env,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await p.communicate()

    if p.returncode != 0:
        raise ValueError(
            f"conda {' '.join(args)} failed, return code {p.returncode}: "
            + stderr.decode()
        )

    return stdout.decode(), stderr.decode()


async def try_get_current_conda_env() -> Optional[str]:
    """
    Returns the result of conda env export on the current conda environment. Returns
    None if there's no currently active conda environment
    """
    env_path = os.environ.get("CONDA_PREFIX")
    if env_path is None:
        return None
    out, _ = await _run(["env", "export", "-p", env_path])
    return out


async def env_export(name_or_path: str) -> str:
    """Runs `conda env export` on the conda environment with the given name, and returns
    the results.

    :param name_or_path: name or full path to the conda environment
    :return: Output of the conda env export command.
    """
    out, _ = await _run(["env", "list", "--json"])
    result = json.loads(out)
    envs: List[str] = result["envs"]
    if len(envs) == 0:
        raise ValueError("No conda environments found")
    elif len(envs) == 1:
        # praise conda: the first path is always the base path
        if name_or_path != "base":
            raise ValueError(
                f"Conda environment {name_or_path} not found - only base environment is"
                " defined"
            )
        env_path = envs[0]
    else:
        if name_or_path in envs:
            env_path = name_or_path
        else:
            for env in envs:
                if name_or_path in env:
                    env_path = env
                    break
            else:
                raise ValueError(f"Conda environment {name_or_path} not found.")

    # TODO: see https://github.com/conda/conda/issues/5253
    # there is also conda list --explicit/--export
    out, _ = await _run(["env", "export", "-p", env_path])
    return out


_CONDA_ENVIRONMENT_TIMEOUT = 30 * 60


async def get_cached_or_create_conda_environment(
    environment_hash: str,
    environment_yml_file_path: str,
    prerequisites: EnvironmentSpecPrerequisites,
    new_environment_path: str,
    try_get_file: Callable[[str, str], Awaitable[bool]],
    upload_file: Callable[[str, str], Awaitable[None]],
) -> str:
    """
    If the desired conda environment exists, does nothing. If the environment has been
    cached, creates it from the cache. Otherwise creates the environment from scratch
    and caches it. Returns the path to the newly created python interpreter.

    try_get_file and upload_file are for interacting with the cache, see
    compile_environment_spec_locally for more details
    """
    new_environment_interpreter = os.path.join(new_environment_path, "bin", "python")

    with filelock.FileLock(f"{new_environment_path}.lock", _CONDA_ENVIRONMENT_TIMEOUT):
        if os.path.exists(new_environment_path):
            return new_environment_interpreter

        remote_cached_file_name = f"{STORAGE_ENV_CACHE_PREFIX}{environment_hash}.tar.gz"
        local_cached_file = f"{new_environment_path}.tar.gz"
        download_succeeded = await try_get_file(
            remote_cached_file_name, local_cached_file
        )
        if download_succeeded:
            try:
                print("Unpacking cached conda environment")
                os.makedirs(new_environment_path, exist_ok=True)
                try:
                    # TODO maybe cleaner to use the built-in python tar libraries?
                    return_code = await (
                        await asyncio.create_subprocess_exec(
                            "tar", "-xzf", local_cached_file, "-C", new_environment_path
                        )
                    ).wait()
                    if return_code != 0:
                        raise ValueError(
                            f"Unpacking cached conda environment {local_cached_file} "
                            f"returned code {return_code}"
                        )

                    print(os.listdir(os.path.join(new_environment_path, "bin")))
                    # conda-unpack shouldn't really be necessary because we always
                    # recreate the environment in exactly the same place, but feels
                    # safer just to run it anyways
                    conda_unpack_command = (
                        "source "
                        f"{os.path.join(new_environment_path, 'bin', 'activate')} && "
                        f"{os.path.join(new_environment_path, 'bin', 'conda-unpack')}"
                    )
                    return_code = await (
                        await asyncio.create_subprocess_exec(
                            "/bin/bash",
                            "-c",
                            conda_unpack_command,
                        )
                    ).wait()
                    if return_code != 0:
                        raise ValueError(
                            f"Calling conda-unpack on restored conda environment "
                            f"{new_environment_path} returned code {return_code}"
                        )
                    return new_environment_interpreter
                except BaseException:
                    remove_corrupted_environment(new_environment_path)
                    raise
            finally:
                try:
                    os.remove(local_cached_file)
                except asyncio.CancelledError:
                    raise
                except BaseException:
                    pass

        print("Creating the conda environment")
        try:
            await create_conda_environment(
                environment_yml_file_path, new_environment_path
            )
        except BaseException:
            remove_corrupted_environment(new_environment_path)
            raise

        try:
            import conda_pack  # see note on reference in pyproject.toml
        except ImportError:
            print(
                "Warning unable to cache conda environment because conda_pack is "
                "missing"
            )
            return new_environment_interpreter

        # TODO we shouldn't wait for this to start running the job but we also shouldn't
        # kill the container until this finishes
        print("Caching the conda environment")
        try:
            # there might be an old file from a failed attempt
            os.remove(local_cached_file)
        except OSError:
            pass
        conda_pack.pack(prefix=new_environment_path, output=local_cached_file)
        await upload_file(local_cached_file, remote_cached_file_name)

        return new_environment_interpreter


async def create_conda_environment(
    environment_yml_file_path: str, new_environment_path: str
) -> None:
    # TODO take prerequisites into account, also support files generated by `conda list
    # --explicit` as well as `conda env export`
    # mamba create -n foo3 --no-default-packages

    # this code is roughly equivalent to the code in CondaDockerfile and
    # CondaDockerfile
    return_code = await (
        await asyncio.create_subprocess_exec(
            "/opt/conda/bin/mamba",
            "env",
            "create",
            "-f",
            environment_yml_file_path,
            "-p",
            new_environment_path,
        )
    ).wait()
    if return_code != 0:
        raise ValueError(
            f"Conda environment creation in {new_environment_path} failed with return "
            f"code {return_code}"
        )
